#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Decision utilties for resampling."""

import logging
import os

import yaml
from pyresample.geometry import SwathDefinition
from satpy._config import config_search_paths
from satpy.utils import recursive_dict_update
from satpy.writers import DecisionTree

from polar2grid.utils.legacy_compat import get_sensor_alias

logger = logging.getLogger(__name__)


class ResamplerDecisionTree(DecisionTree):
    """Helper class to determine resampler algorithm and other options."""

    def __init__(self, *decision_dicts, **kwargs):
        """Init the decision tree."""
        match_keys = kwargs.pop(
            "match_keys",
            (
                "name",
                "reader",
                "platform_name",
                "sensor",
                "area_type",
                "standard_name",
                "units",
            ),
        )
        self.prefix = kwargs.pop("config_section", "resampling")
        multival_keys = kwargs.pop("multival_keys", ["sensor"])
        super(ResamplerDecisionTree, self).__init__(decision_dicts, match_keys, multival_keys)

    @classmethod
    def from_configs(cls, config_filename="resampling.yaml"):
        config_paths = config_search_paths(config_filename)
        return cls(*config_paths)

    def add_config_to_tree(self, *config_files):
        """Add configuration to tree."""
        conf = {}
        for config_file in config_files:
            if os.path.isfile(config_file):
                with open(config_file) as fd:
                    resample_config = yaml.load(fd, Loader=yaml.UnsafeLoader)
                    if resample_config is None:
                        # empty file
                        continue
                    resampling_section = resample_config.get(self.prefix, {})
                    if not resampling_section:
                        logging.debug("Config '{}' has no '{}' section or it is empty".format(config_file, self.prefix))
                        continue
                    conf = recursive_dict_update(conf, resampling_section)
            elif isinstance(config_file, dict):
                conf = recursive_dict_update(conf, config_file)
            else:
                logger.debug("Loading resampling config string")
                d = yaml.load(config_file, Loader=yaml.UnsafeLoader)
                if not isinstance(d, dict):
                    raise ValueError("YAML file doesn't exist or string is not YAML dict: {}".format(config_file))
                conf = recursive_dict_update(conf, d)
        self._build_tree(conf)

    def find_match(self, **query_dict):
        """Find a match."""
        query_dict["area_type"] = "swath" if isinstance(query_dict["area"], SwathDefinition) else "area"
        query_dict["sensor"] = get_sensor_alias(query_dict.get("sensor"))

        try:
            return super().find_match(**query_dict)
        except KeyError:
            # give a more understandable error message
            raise KeyError(f"No resampling configuration found for {query_dict['area_type']=} | {query_dict['name']=}")



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2013-2021 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Helper script for creating valid grid configuration entries.

The user provides a center longitude and latitude, along with other grid
parameters, and a configuration line is printed to stdout that can be added
to a user's own grid configuration file.

"""
from __future__ import annotations

import os
import sys
import warnings

from pyproj import CRS, Proj

from polar2grid.utils.convert_grids_conf_to_yaml import crs_to_proj_dict, ordered_dump

# grid_name,proj4, proj4_str,width,height,pixel_size_x,pixel_size_y,origin_x,origin_y
CONFIG_LINE_FORMAT = (
    "%(grid_name)s, proj4, %(proj4_str)s, %(width)d, %(height)d, %(pixel_size_x)0.5f, "
    "%(pixel_size_y)0.5f, %(origin_x)s, %(origin_y)s"
)


def determine_projection(center_lon: float, center_lat: float, proj4_str: str = None) -> CRS:
    """Get the 'best' projection to be used based on the center longitude and latitude provided."""
    abs_lat = abs(center_lat)
    if proj4_str is None:
        if abs_lat < 15:
            proj4_str = (
                "+proj=eqc +datum=WGS84 +ellps=WGS84 +lat_ts=%(center_lat)0.5f +lon_0=%(center_lon)0.5f "
                "+units=m +no_defs"
            )
        elif abs_lat < 70:
            proj4_str = (
                "+proj=lcc +datum=WGS84 +ellps=WGS84 +lat_0=%(center_lat)0.5f +lat_1=%(center_lat)0.5f "
                "+lon_0=%(center_lon)0.5f +units=m +no_defs"
            )
        else:
            proj4_str = (
                "+proj=stere +datum=WGS84 +ellps=WGS84 +lat_0=90 +lat_ts=%(center_lat)0.5f "
                "+lon_0=%(center_lon)0.5f +units=m"
            )

    proj4_str = proj4_str % dict(center_lon=center_lon, center_lat=center_lat)
    return CRS.from_user_input(proj4_str)


def get_parser():
    from argparse import SUPPRESS, ArgumentParser

    prog = os.getenv("PROG_NAME", sys.argv[0])
    description = """This script is meant to help those unfamiliar with PROJ.4 and projections
in general. By providing a few grid parameters this script will provide a
grid configuration line that can be added to a user's custom grid
configuration. Based on a center longitude and latitude, the script will
choose an appropriate projection."""
    parser = ArgumentParser(prog=prog, description=description)
    parser.add_argument("grid_name", type=str, help="Unique grid name")
    parser.add_argument("center_longitude", type=float, help="Decimal longitude value for center of grid (-180 to 180)")
    parser.add_argument("center_latitude", type=float, help="Decimal latitude value for center of grid (-90 to 90)")
    parser.add_argument(
        "pixel_size_x",
        type=float,
        help="""Size of each pixel in the X direction in grid units,
meters for default projections.""",
    )
    parser.add_argument(
        "pixel_size_y",
        type=float,
        help="""Size of each pixel in the Y direction in grid units,
meters for default projections.""",
    )
    parser.add_argument("grid_width", type=int, help="Grid width in number of pixels")
    parser.add_argument("grid_height", type=int, help="Grid height in number of pixels")
    parser.add_argument("-p", dest="proj_str", default=None, help="PROJ.4 projection string to override the default")
    # Don't force Y pixel size to be negative (for expert use only)
    parser.add_argument(
        "--dont-touch-ysize", dest="dont_touch_ysize", action="store_true", default=False, help=SUPPRESS
    )
    parser.add_argument("--legacy-format", action="store_true", help="Produce a legacy '.conf' format grid definition.")
    return parser


def _create_legacy_line(
    grid_name: str,
    crs: CRS,
    clon: float,
    clat: float,
    grid_width: int,
    grid_height: int,
    pixel_size_x: float,
    pixel_size_y: float,
) -> str:
    p = Proj(crs)
    origin_x = p(clon, clat)[0] + (grid_width / 2.0 * pixel_size_x * -1)
    origin_y = p(clon, clat)[1] + (grid_height / 2.0 * pixel_size_y * -1)
    if p.crs.is_geographic:
        # Origin is in degrees so we need to add a unit string to it
        origin_x = "%0.5fdeg" % origin_x
        origin_y = "%0.5fdeg" % origin_y
    else:
        origin_lon, origin_lat = p(origin_x, origin_y, inverse=True)
        origin_x = "%0.5fdeg" % origin_lon
        origin_y = "%0.5fdeg" % origin_lat

    valid_config_line = CONFIG_LINE_FORMAT % {
        "grid_name": grid_name,
        "proj4_str": crs.to_proj4(),
        "origin_x": origin_x,
        "origin_y": origin_y,
        "pixel_size_x": pixel_size_x,
        "pixel_size_y": pixel_size_y,
        "width": grid_width,
        "height": grid_height,
    }
    return valid_config_line


def _create_yaml_entry(
    grid_name: str,
    crs: CRS,
    clon: float,
    clat: float,
    grid_width: int,
    grid_height: int,
    pixel_size_x: float,
    pixel_size_y: float,
):
    proj_dict = crs_to_proj_dict(crs)
    area_dict = {
        "projection": proj_dict,
        "shape": {"height": grid_height, "width": grid_width},
        "center": {"x": clon, "y": clat, "units": "degrees"},
        "resolution": {"dx": abs(pixel_size_x), "dy": abs(pixel_size_y)},
    }
    return {grid_name: area_dict}


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = get_parser()
    args = parser.parse_args(argv)

    grid_name = args.grid_name
    clon = args.center_longitude
    clat = args.center_latitude
    pixel_size_x = args.pixel_size_x
    pixel_size_y = args.pixel_size_y
    if pixel_size_y > 0 and not args.dont_touch_ysize:
        pixel_size_y *= -1
    grid_width = args.grid_width
    grid_height = args.grid_height
    crs = determine_projection(clon, clat, args.proj_str)
    warnings.filterwarnings("ignore", module="pyproj", category=UserWarning)

    if clon < -180 or clon > 180:
        print("ERROR: Center longitude must be between -180 and 180 degrees")
        return -1
    if clat < -90 or clat > 90:
        print("ERROR: Center latitude must be between -90 and 90 degrees")
        return -1

    if args.legacy_format:
        valid_config_line = _create_legacy_line(
            grid_name,
            crs,
            clon,
            clat,
            grid_width,
            grid_height,
            pixel_size_x,
            pixel_size_y,
        )
        print(valid_config_line)
    else:
        yaml_dict = _create_yaml_entry(
            grid_name,
            crs,
            clon,
            clat,
            grid_width,
            grid_height,
            pixel_size_x,
            pixel_size_y,
        )
        yml_str = ordered_dump(yaml_dict, default_flow_style=None)
        print(yml_str)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2022 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Test fixtures representing AVHRR swath data."""
from __future__ import annotations

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyresample import SwathDefinition
from satpy import Scene
from satpy.tests.utils import make_dataid

from ._fixture_utils import START_TIME, _TestingScene, generate_lonlat_data

AVHRR_SHAPE = (2361, 2048)
AVHRR_CHUNKS = AVHRR_SHAPE

AVHRR_IDS = [
    make_dataid(name="1", wavelength=(0.58, 0.63, 0.68), resolution=1050, calibration="reflectance"),
]


@pytest.fixture(scope="session")
def avhrr_l1b_swath_def() -> SwathDefinition:
    lons, lats = generate_lonlat_data(AVHRR_SHAPE)
    lons_data_arr = xr.DataArray(
        da.from_array(lons, chunks=AVHRR_CHUNKS),
        dims=("y", "x"),
        attrs={
            "platform_name": "noaa19",
            "sensor": "avhrr-3",
            "start_time": START_TIME,
            "end_time": START_TIME,
            "resolution": 1050,
            "reader": "avhrr_l1b_aapp",
        },
    )
    lats_data_arr = xr.DataArray(
        da.from_array(lats, chunks=AVHRR_CHUNKS),
        dims=("y", "x"),
        attrs={
            "platform_name": "noaa19",
            "sensor": "avhrr-3",
            "start_time": START_TIME,
            "end_time": START_TIME,
            "resolution": 1050,
            "reader": "avhrr_l1b_aapp",
        },
    )
    return SwathDefinition(lons_data_arr, lats_data_arr)


@pytest.fixture
def avhrr_l1b_1_data_array(avhrr_l1b_swath_def) -> xr.DataArray:
    return xr.DataArray(
        da.zeros(AVHRR_SHAPE, dtype=np.float32),
        dims=("y", "x"),
        attrs={
            "area": avhrr_l1b_swath_def,
            "platform_name": "noaa19",
            "sensor": "avhrr-3",
            "name": "1",
            "start_time": START_TIME,
            "end_time": START_TIME,
            "resolution": 1050,
            "reader": "avhrr_l1b_aapp",
            "calibration": "reflectance",
            "standard_name": "toa_bidirectional_reflectance",
            "units": "%",
        },
    )


@pytest.fixture
def avhrr_l1b_1_scene(avhrr_l1b_1_data_array) -> Scene:
    scn = _TestingScene(
        reader="avhrr_l1b_aapp",
        filenames=["/fake/filename"],
        data_array_dict={
            AVHRR_IDS[0]: avhrr_l1b_1_data_array.copy(),
        },
        all_dataset_ids=AVHRR_IDS,
        available_dataset_ids=AVHRR_IDS[:1],
    )
    return scn



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Helper functions for resampling Satpy Scenes."""

from __future__ import annotations

import logging
import os
from typing import List, Optional, Union

import numpy as np
from pyproj import Proj
from pyresample import parse_area_file
from pyresample.geometry import AreaDefinition, DynamicAreaDefinition
from satpy import Scene
from satpy.resample import get_area_def

from polar2grid.filters.resample_coverage import ResampleCoverageFilter
from polar2grid.grids import GridManager

from ..filters._utils import PRGeometry
from .resample_decisions import ResamplerDecisionTree

logger = logging.getLogger(__name__)

# TypeAlias
AreaSpecifier = Union[AreaDefinition, str, None]
ListOfAreas = List[Union[AreaDefinition, str, None]]

GRIDS_YAML_FILEPATH = os.path.realpath(os.path.join(os.path.dirname(__file__), "..", "grids", "grids.yaml"))


def _crs_equal(a, b):
    """Compare two projection dictionaries for "close enough" equality."""
    # pyproj 2.0+
    if hasattr(a, "crs") and hasattr(b, "crs"):
        return a.crs == b.crs

    # fallback
    from osgeo import osr

    a = dict(sorted(a.proj_dict.items()))
    b = dict(sorted(b.proj_dict.items()))
    p1 = Proj(a)
    p2 = Proj(b)
    s1 = osr.SpatialReference()
    s1.ImportFromProj4(p1.srs)
    s2 = osr.SpatialReference()
    s2.ImportFromProj4(p2.srs)
    return s1.IsSame(s2)


def _get_preserve_resolution(preserve_resolution, resampler, areas_to_resample):
    """Determine if we should preserve native resolution products.

    Preserving native resolution should only happen if:

    1. The 'native' resampler is used
    2. At least one of the areas provided to resampling are 'MIN' or 'MAX'
    3. The user didn't ask to *not* preserve it.

    """
    # save original native resolution if possible
    any_minmax = any(x in ["MIN", "MAX"] for x in areas_to_resample)
    is_native = resampler == "native"
    is_default = resampler is None
    return any_minmax and (is_native or is_default) and preserve_resolution


def _get_legacy_and_yaml_areas(grid_configs: list[str, ...]) -> tuple[GridManager, dict[str, AreaDefinition]]:
    if "grids.conf" in grid_configs:
        logger.debug("Replacing 'grids.conf' with builtin YAML grid configuration file.")
        grid_configs[grid_configs.index("grids.conf")] = GRIDS_YAML_FILEPATH
    if not grid_configs:
        grid_configs = [GRIDS_YAML_FILEPATH]
    p2g_grid_configs = [x for x in grid_configs if x.endswith(".conf")]
    pyresample_area_configs = [x for x in grid_configs if not x.endswith(".conf")]
    if p2g_grid_configs:
        grid_manager = GridManager(*p2g_grid_configs)
    else:
        grid_manager = {}

    if pyresample_area_configs:
        yaml_areas = parse_area_file(pyresample_area_configs)
        yaml_areas = {x.area_id: x for x in yaml_areas}
    else:
        yaml_areas = {}

    return grid_manager, yaml_areas


def _get_area_def_from_name(
    area_name: Optional[str], input_scene: Scene, grid_manager: GridManager, yaml_areas: list
) -> Optional[PRGeometry]:
    if area_name is None:
        # no resampling
        area_def = None
    elif area_name == "MAX":
        area_def = input_scene.finest_area()
    elif area_name == "MIN":
        area_def = input_scene.coarsest_area()
    elif area_name in yaml_areas:
        area_def = yaml_areas[area_name]
    elif area_name in grid_manager:
        p2g_def = grid_manager[area_name]
        area_def = p2g_def.to_satpy_area()
    else:
        # get satpy builtin area
        area_def = get_area_def(area_name)
    return area_def


class AreaDefResolver:
    def __init__(self, input_scene, grid_configs):
        grid_manager, yaml_areas = _get_legacy_and_yaml_areas(grid_configs)
        self.input_scene = input_scene
        self.grid_manager = grid_manager
        self.yaml_areas = yaml_areas

    def has_dynamic_extents(self, area_name: Optional[str]) -> bool:
        area_def = self[area_name]
        is_dynamic = isinstance(area_def, DynamicAreaDefinition)
        return is_dynamic and area_def.area_extent is None

    def __getitem__(self, area_name: Optional[str]) -> Optional[PRGeometry]:
        area_def = _get_area_def_from_name(area_name, self.input_scene, self.grid_manager, self.yaml_areas)
        return area_def

    def get_frozen_area(self, area_name: Optional[str], **kwargs) -> Optional[PRGeometry]:
        area_def = self[area_name]
        return self._freeze_area_if_dynamic(area_def, **kwargs)

    def _freeze_area_if_dynamic(self, area_def: PRGeometry, **kwargs) -> PRGeometry:
        if isinstance(area_def, DynamicAreaDefinition):
            logger.info("Computing dynamic grid parameters...")
            area_def = area_def.freeze(self.input_scene.finest_area(), **kwargs)
            logger.debug("Frozen dynamic area: %s", area_def)
        return area_def


def resample_scene(
    input_scene: Scene,
    areas_to_resample: ListOfAreas,
    grid_configs: list[str, ...],
    resampler: Optional[str],
    antimeridian_mode: str = "modify_crs",
    preserve_resolution: bool = True,
    grid_coverage: Optional[float] = None,
    is_polar2grid: bool = True,
    **resample_kwargs,
) -> list[tuple[Scene, set]]:
    """Resample a single Scene to multiple target areas."""
    area_resolver = AreaDefResolver(input_scene, grid_configs)
    resampling_groups = _get_groups_to_resample(resampler, input_scene, is_polar2grid, resample_kwargs)
    wishlist: set = input_scene.wishlist.copy()
    scenes_to_save = []
    for (resampler, _resample_kwargs, default_target), data_ids in resampling_groups.items():
        areas = _areas_to_resample(areas_to_resample, resampler, default_target)
        scene_to_resample: Scene = input_scene.copy(datasets=data_ids)
        preserve_resolution = _get_preserve_resolution(preserve_resolution, resampler, areas)
        preserved_products = _products_to_preserve_resolution(preserve_resolution, wishlist, scene_to_resample)
        if preserved_products:
            scenes_to_save.append((scene_to_resample, preserved_products))

        logger.debug("Products to preserve resolution for: {}".format(preserved_products))
        logger.debug("Products to use new resolution for: {}".format(set(wishlist) - preserved_products))
        # convert hashable tuple to dict
        _resample_kwargs = _redict_hashable_kwargs(_resample_kwargs)
        _grid_cov = _resample_kwargs.get("grid_coverage", grid_coverage)
        if _grid_cov is None:
            _grid_cov = 0.1
        for area_name in areas:
            area_def = area_resolver.get_frozen_area(area_name, antimeridian_mode=antimeridian_mode)
            has_dynamic_extents = area_resolver.has_dynamic_extents(area_name)
            rs = _get_default_resampler(resampler, area_name, area_def, input_scene)
            new_scn = _filter_and_resample_scene_to_single_area(
                area_name,
                area_def,
                _grid_cov,
                has_dynamic_extents,
                scene_to_resample,
                data_ids,
                rs,
                _resample_kwargs,
                preserve_resolution,
            )
            if new_scn is None:
                continue

            # we only want to try to save products that we asked for and that
            # we were actually able to generate. Composite generation may have
            # modified the original DataID so we can't use
            # 'resampled_products'.
            _resampled_products = (new_scn.wishlist & set(new_scn.keys())) - preserved_products
            if _resampled_products:
                scenes_to_save.append((new_scn, _resampled_products))

    return scenes_to_save


def _get_groups_to_resample(
    resampler: str,
    input_scene: Scene,
    is_polar2grid: bool,
    user_resample_kwargs: dict,
) -> dict:
    resampling_dtree = ResamplerDecisionTree.from_configs()
    resampling_groups = {}
    for data_id in input_scene.keys():
        resampling_args = resampling_dtree.find_match(**input_scene[data_id].attrs)
        default_resampler = resampling_args.get("resampler")
        resampler_kwargs = resampling_args.get("kwargs", {}).copy()
        resampler_kwargs.update(user_resample_kwargs)
        default_target = resampling_args.get("default_target", None)
        resampler = resampler if resampler is not None else default_resampler
        if default_target is None:
            default_target = _default_grid(resampler, is_polar2grid)
        hashable_kwargs = _hashable_kwargs(resampler_kwargs)
        resampling_groups.setdefault((resampler, hashable_kwargs, default_target), []).append(data_id)
    return resampling_groups


def _default_grid(resampler, is_polar2grid):
    if resampler in [None, "native"]:
        default_target = "MAX"
    else:
        default_target = "wgs84_fit" if is_polar2grid else "MAX"
    return default_target


def _hashable_kwargs(kwargs):
    return tuple(sorted(kwargs.items()))


def _redict_hashable_kwargs(kwargs_tuple):
    return dict(kwargs_tuple)


def _get_default_resampler(resampler, area_name, area_def, input_scene):
    if resampler is None and area_def is not None:
        rs = (
            "native"
            if area_name in ["MIN", "MAX"] or _is_native_grid(area_def, input_scene.finest_area())
            else "nearest"
        )
        logger.debug("Setting default resampling to '{}' for grid '{}'".format(rs, area_name))
    else:
        rs = resampler
    return rs


def _filter_and_resample_scene_to_single_area(
    area_name: str,
    area_def: Optional[PRGeometry],
    grid_coverage: float,
    has_dynamic_extents: bool,
    input_scene: Scene,
    data_ids_to_resample: list,
    rs: str,
    resample_kwargs: dict,
    preserve_resolution: bool,
) -> Optional[Scene]:
    filtered_data_ids, filtered_scn = _filter_scene_with_grid_coverage(
        area_name,
        area_def,
        rs,
        grid_coverage,
        has_dynamic_extents,
        input_scene,
        data_ids_to_resample,
    )
    if filtered_scn is None:
        return
    new_scn = _resample_scene_to_single_area(
        filtered_scn,
        area_name,
        area_def,
        rs,
        filtered_data_ids,
        resample_kwargs,
        preserve_resolution,
    )
    return new_scn


def _is_native_grid(grid, max_native_area):
    """Check if the current grid is in the native data projection."""
    if not isinstance(max_native_area, AreaDefinition):
        return False
    if not isinstance(grid, AreaDefinition):
        return False
    if not _crs_equal(max_native_area, grid):
        return False
    if not np.allclose(np.array(max_native_area.area_extent), np.array(grid.area_extent)):
        return False
    if max_native_area.width < grid.width:
        return (grid.width / max_native_area.width).is_integer()
    else:
        return (max_native_area.width / grid.width).is_integer()


def _filter_scene_with_grid_coverage(
    area_name: str,
    area_def: PRGeometry,
    resampler: str,
    coverage_threshold: float,
    has_dynamic_extents: bool,
    scene_to_resample: Scene,
    data_ids: list,
):
    if area_def is not None and resampler != "native" and coverage_threshold > 0.0 and not has_dynamic_extents:
        logger.info("Checking products for sufficient output grid coverage (grid: '%s')...", area_name)
        filter = ResampleCoverageFilter(target_area=area_def, coverage_fraction=coverage_threshold)
        scene_to_resample = filter.filter_scene(scene_to_resample)
        if scene_to_resample is None:
            logger.warning("No products were found to overlap with '%s' grid.", area_name)
            return None, None
        if data_ids is not None:
            data_ids = list(set(data_ids) & set(scene_to_resample.keys()))
    return data_ids, scene_to_resample


def _resample_scene_to_single_area(
    scene_to_resample: Scene,
    area_name: str,
    area_def: PRGeometry,
    rs: str,
    data_ids: list,
    resample_kwargs: dict,
    preserve_resolution: bool,
) -> Optional[Scene]:
    if area_def is not None:
        logger.info("Resampling to '%s' using '%s' resampling...", area_name, rs)
        logger.debug("Resampling to '%s' using resampler '%s' with %s", area_name, rs, resample_kwargs)
        new_scn = scene_to_resample.resample(area_def, resampler=rs, datasets=data_ids, **resample_kwargs)
    elif not preserve_resolution:
        # the user didn't want to resample to any areas
        # the user also requested that we don't preserve resolution
        # which means we have to save this Scene's datasets
        # because they won't be saved
        new_scn = scene_to_resample
    else:
        # No resampling and any preserved resolution datasets were saved earlier
        return
    return new_scn


def _areas_to_resample(
    areas_to_resample: Optional[ListOfAreas], resampler: Optional[str], default_target: Optional[AreaSpecifier]
) -> ListOfAreas:
    areas = areas_to_resample
    if areas is None:
        if resampler in ["native"]:
            logging.debug("Using default resampling target area 'MAX'.")
            areas = ["MAX"]
        elif default_target is None:
            raise ValueError("No destination grid/area specified and no default available (use -g flag).")
        else:
            logging.debug("Using default resampling target area '%s'.", default_target)
            areas = [default_target]
    elif not areas:
        areas = [None]

    return areas


def _products_to_preserve_resolution(preserve_resolution, wishlist, scene_to_resample):
    if preserve_resolution:
        return set(wishlist) & set(scene_to_resample.keys())
    return set()



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
#     input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""The AVHRR reader is for reading AAPP L1B files for the AVHRR instrument.

These files are a custom binary format. The reader can be specified with
the ``polar2grid.sh`` command using the ``avhrr`` or ``avhrr_l1b_aapp`` reader
name.

This reader's default resampling algorithm is ``ewa`` for Elliptical Weighted
Averaging resampling. The ``--weight-delta-max`` option is set to 10 and
``--weight-distance-max`` is set to 1.

The AVHRR reader provides the following products:

+--------------------+--------------------------------------------+
| Product Name       | Description                                |
+====================+============================================+
| band1_vis          | Band 1 Visible                             |
+--------------------+--------------------------------------------+
| band2_vis          | Band 2 Visible                             |
+--------------------+--------------------------------------------+
| band3a_vis         | Band 3A Visible                            |
+--------------------+--------------------------------------------+
| band3b_bt          | Band 3B Brightness Temperature             |
+--------------------+--------------------------------------------+
| band4_vis          | Band 4 Brightness Temperature              |
+--------------------+--------------------------------------------+
| band5_vis          | Band 5 Brightness Temperature              |
+--------------------+--------------------------------------------+

"""
from __future__ import annotations

from argparse import ArgumentParser, _ArgumentGroup
from typing import Optional

from satpy import DataQuery

from ._base import ReaderProxyBase

FILTERS = {
    "day_only": {
        "standard_name": [
            "toa_bidirectional_reflectance",
            "corrected_reflectance",
        ],
    },
}

VIS_PRODUCTS = [
    "band1_vis",
    "band2_vis",
    "band3a_vis",
]

IR_PRODUCTS = [
    "band3b_bt",
    "band4_bt",
    "band5_bt",
]

PRODUCT_ALIASES = {
    "band1_vis": DataQuery(name="1", calibration="reflectance"),
    "band2_vis": DataQuery(name="2", calibration="reflectance"),
    "band3a_vis": DataQuery(name="3a", calibration="reflectance"),
    "band3b_bt": DataQuery(name="3b", calibration="brightness_temperature"),
    "band4_bt": DataQuery(name="4", calibration="brightness_temperature"),
    "band5_bt": DataQuery(name="5", calibration="brightness_temperature"),
}


class ReaderProxy(ReaderProxyBase):
    """Provide Polar2Grid-specific information about this reader's products."""

    is_polar2grid_reader = True

    def get_default_products(self) -> list[str]:
        """Get products to load if users hasn't specified any others."""
        return VIS_PRODUCTS + IR_PRODUCTS

    def get_all_products(self) -> list[str]:
        """Get all polar2grid products that could be loaded."""
        return VIS_PRODUCTS + IR_PRODUCTS

    @property
    def _aliases(self) -> dict[str, DataQuery]:
        return PRODUCT_ALIASES


def add_reader_argument_groups(
    parser: ArgumentParser, group: Optional[_ArgumentGroup] = None
) -> tuple[Optional[_ArgumentGroup], Optional[_ArgumentGroup]]:
    """Add reader-specific command line arguments to an existing argument parser.

    If ``group`` is provided then arguments are added to this group. If not,
    a new group is added to the parser and arguments added to this new group.

    """
    if group is None:
        group = parser.add_argument_group(title="AVHRR L1b AAPP Reader")
    return group, None



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2016-2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""AMSR2 L1B files contain various parameters from the GCOM-W1 AMSR2 instrument.

This reader can be used by specifying the reader name
``amsr2_l1b`` to the ``polar2grid.sh`` script.
Supported files usually have the following naming scheme::

    GW1AM2_201607201808_128A_L1DLBTBR_1110110.h5

This reader's default remapping algorithm is ``nearest`` for nearest
neighbor resampling due to the instruments scan pattern and swath shape.
The ``--distance_upper_bound`` flag defaults to 12.

Currently this reader provides only the following datasets:

+---------------------------+-----------------------------------------------------------+
| Product Name              | Description                                               |
+===========================+===========================================================+
| btemp_36.5v               | Brightness Temperature 36.5GHz Polarization Vertical      |
+---------------------------+-----------------------------------------------------------+
| btemp_36.5h               | Brightness Temperature 36.5GHz Polarization Horizontal    |
+---------------------------+-----------------------------------------------------------+
| btemp_89.0av              | Brightness Temperature 89.0GHz A Polarization Vertical    |
+---------------------------+-----------------------------------------------------------+
| btemp_89.0ah              | Brightness Temperature 89.0GHz A Polarization Horizontal  |
+---------------------------+-----------------------------------------------------------+
| btemp_89.0bv              | Brightness Temperature 89.0GHz B Polarization Vertical    |
+---------------------------+-----------------------------------------------------------+
| btemp_89.0bh              | Brightness Temperature 89.0GHz B Polarization Horizontal  |
+---------------------------+-----------------------------------------------------------+

Special AMSR2 Naval Research Lab (NRL) PNG Scaling
--------------------------------------------------

A common use case for the AMSR2 L1B reader is to generate PNG images similar
to those generated by the U.S. Naval Research Lab (NRL) with a colormap and
coastlines. This requires using an alternate non-default scaling configuration
provided in the tarball. It can be used by providing the
``--extra-config-path $POLAR2GRID_HOME/example_enhancements/amsr2_png``
flag when generating AMSR2 L1B GeoTIFFs. This allows the scaling provided
in the ``$POLAR2GRID_HOME/example_enhancements/amsr2_png/enhancements/generic.yaml``
file to be used.  Once this rescaling has been done, colormap files can be
found in ``$POLAR2GRID_HOME/colormaps`` which can then be applied using the
the `add_colormap.sh` script.

See the example section :doc:`../../examples/amsr2_example`
for more information on generating these NRL-like PNGs.

"""
from __future__ import annotations

from argparse import ArgumentParser, _ArgumentGroup
from typing import Optional

from satpy import DataQuery

from ._base import ReaderProxyBase

DEFAULT_CHANNELS = [
    # "btemp_10.7v",
    # "btemp_10.7h",
    "btemp_36.5v",
    "btemp_36.5h",
    "btemp_89.0av",
    "btemp_89.0ah",
    "btemp_89.0bv",
    "btemp_89.0bh",
]


class ReaderProxy(ReaderProxyBase):
    """Provide Polar2Grid-specific information about this reader's products."""

    is_polar2grid_reader = True

    def get_default_products(self) -> list[str]:
        """Get products to load if users hasn't specified any others."""
        return DEFAULT_CHANNELS

    def get_all_products(self) -> list[str]:
        """Get all polar2grid products that could be loaded."""
        return DEFAULT_CHANNELS

    @property
    def _aliases(self) -> dict[str, DataQuery]:
        return {}


def add_reader_argument_groups(
    parser: ArgumentParser, group: Optional[_ArgumentGroup] = None
) -> tuple[Optional[_ArgumentGroup], Optional[_ArgumentGroup]]:
    """Add reader-specific command line arguments to an existing argument parser.

    If ``group`` is provided then arguments are added to this group. If not,
    a new group is added to the parser and arguments added to this new group.

    """
    if group is None:
        group = parser.add_argument_group(title="AMSR2 L1B Reader")
    return group, None



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2022 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Tests day/night filtering."""
import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyresample import SwathDefinition
from satpy import Scene

from polar2grid.filters.day_night import DayCoverageFilter, NightCoverageFilter


@pytest.mark.parametrize(
    "filter_cls",
    [DayCoverageFilter, NightCoverageFilter],
)
@pytest.mark.parametrize(
    "criteria",
    ["default", None, {}],
)
def test_daynight_filter_no_filter_check_cases(filter_cls, criteria, viirs_sdr_i01_data_array):
    scn = Scene()
    scn["I01"] = viirs_sdr_i01_data_array

    if criteria == "default":
        filter = filter_cls()
    else:
        filter = filter_cls(criteria)
    new_scn = filter.filter_scene(scn)
    assert new_scn is not None
    assert new_scn is not scn
    assert "I01" in new_scn


@pytest.mark.parametrize(
    ("filter_cls", "kwargs"),
    [
        (DayCoverageFilter, {"day_fraction": 0.95}),
        (NightCoverageFilter, {"night_fraction": 0.5}),
    ],
)
@pytest.mark.parametrize("add_nans", [False, True])
def test_daynight_filter_filter_cases(filter_cls, add_nans, kwargs, viirs_sdr_i01_data_array):
    scn = Scene()
    scn["I01"] = viirs_sdr_i01_data_array  # calculates to ~0.8 day
    if add_nans:
        _replace_swath_def_with_some_bounding_nans(scn["I01"])

    criteria = {"standard_name": ["toa_bidirectional_reflectance"]}
    filter = filter_cls(criteria, **kwargs)
    new_scn = filter.filter_scene(scn)
    assert new_scn is None  # all filtered
    assert "I01" in scn


def _replace_swath_def_with_some_bounding_nans(data_arr: xr.DataArray) -> None:
    swath_def = data_arr.attrs["area"]
    lons = swath_def.lons
    lats = swath_def.lats
    lons_np = lons.compute()
    lons_np.data[: lons_np.shape[0], -1] = np.nan
    new_lons = xr.DataArray(
        da.from_array(lons_np.data, chunks=lons.chunks),
        attrs=lons.attrs.copy(),
    )
    new_swath_def = SwathDefinition(new_lons, lats)
    data_arr.attrs["swath_def"] = new_swath_def


def test_daynight_filter_basic_composite_case(viirs_sdr_i04_data_array):
    """Test that composites are removed from the wishlist and not recomputed after resampling.

    See https://github.com/ssec/polar2grid/issues/544.

    """
    from polar2grid.utils.config import add_polar2grid_config_paths

    add_polar2grid_config_paths()

    scn = Scene()
    scn["I04"] = viirs_sdr_i04_data_array
    i05_data_arr = viirs_sdr_i04_data_array.copy(deep=True)
    i05_data_arr.attrs["name"] = "I05"
    scn["I05"] = i05_data_arr
    scn.load(["ifog"])

    filter = NightCoverageFilter({"standard_name": "temperature_difference"}, night_fraction=0.2)
    new_scn = filter.filter_scene(scn)
    assert new_scn is not None
    assert "ifog" not in new_scn

    resampled_scn = new_scn.resample(resampler="native")
    assert "ifog" not in resampled_scn



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2020-2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""The AMSR-2 L2 GAASP reader is for reading NetCDF4 from the NOAA GAASP software.

+-----------------------------------+--------------------------------------------+
| Product Name                      | Description                                |
+===========================+====================================================+
| Rain_Rate                         | Surface Rain Rate                          |
+-----------------------------------+--------------------------------------------+
| CLW                               | Cloud Liquid Water                         |
+-----------------------------------+--------------------------------------------+
| SST                               | Sea Surface Temperature                    |
+-----------------------------------+--------------------------------------------+
| TPW                               | Total Precipitable Water                   |
+-----------------------------------+--------------------------------------------+
| WSPD                              | Wind Speed                                 |
+-----------------------------------+--------------------------------------------+
| Snow_Cover                        | Snow Cover (Snow Surface Flag)             |
+-----------------------------------+--------------------------------------------+
| SWE                               | Snow Water Equivalent                      |
+-----------------------------------+--------------------------------------------+
| Snow_Depth                        | Snow Depth                                 |
+-----------------------------------+--------------------------------------------+
| Soil_Moisture                     | Soil Moisture                              |
+-----------------------------------+--------------------------------------------+
| NASA_Team_2_Ice_Concentration_NH  |  Ice Concentration (Northern Hemisphere)   |
+-----------------------------------+--------------------------------------------+

"""
from __future__ import annotations

from argparse import ArgumentParser, _ArgumentGroup
from typing import Optional

from satpy import DataQuery

from ._base import ReaderProxyBase

OCEAN_PRECIP_PRODUCTS = [
    "Rain_Rate",
    "CLW",
    "SST",
    "TPW",
    "WSPD",
]
SNOW_PRODUCTS = [
    "Snow_Cover",
    "SWE",
    "Snow_Depth",
]
SOIL_PRODUCTS = [
    "Soil_Moisture",
]
SEAICE_PRODUCTS = [
    "NASA_Team_2_Ice_Concentration_NH",
    # "Latency_NH",
]

DEFAULT_PRODUCTS = OCEAN_PRECIP_PRODUCTS + SNOW_PRODUCTS + SOIL_PRODUCTS + SEAICE_PRODUCTS


class ReaderProxy(ReaderProxyBase):
    """Provide Polar2Grid-specific information about this reader's products."""

    is_polar2grid_reader = True

    def get_default_products(self) -> list[str]:
        """Get products to load if users hasn't specified any others."""
        return DEFAULT_PRODUCTS

    def get_all_products(self) -> list[str]:
        """Get all polar2grid products that could be loaded."""
        return DEFAULT_PRODUCTS

    @property
    def _aliases(self) -> dict[str, DataQuery]:
        return {}


def add_reader_argument_groups(
    parser: ArgumentParser, group: Optional[_ArgumentGroup] = None
) -> tuple[Optional[_ArgumentGroup], Optional[_ArgumentGroup]]:
    """Add reader-specific command line arguments to an existing argument parser.

    If ``group`` is provided then arguments are added to this group. If not,
    a new group is added to the parser and arguments added to this new group.

    """
    if group is None:
        group = parser.add_argument_group(title="AMSR2 L2 GAASP Reader")
    return group, None



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2020 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""The VIIRS SDR Reader operates on Sensor Data Record (SDR) HDF5 files from
the Suomi National Polar-orbiting Partnership's (NPP) and/or the NOAA20
Visible/Infrared Imager Radiometer Suite (VIIRS) instrument. The VIIRS
SDR reader ignores filenames and uses internal file content to determine
the type of file being provided, but SDR are typically named as below
and have corresponding geolocation files::

    SVI01_npp_d20120225_t1801245_e1802487_b01708_c20120226002130255476_noaa_ops.h5

The VIIRS SDR reader supports all instrument spectral bands, identified as
the products shown below. It supports terrain corrected or non-terrain corrected
navigation files. Geolocation files must be included when specifying filepaths to
readers and ``polar2grid.sh``. The VIIRS reader can be specified to the ``polar2grid.sh`` script
with the reader name ``viirs_sdr``.

This reader's default remapping algorithm is ``ewa`` for Elliptical Weighted
Averaging resampling. The ``--weight-delta-max`` parameter set to 40 and the
``--weight-distance-max`` parameter set to 2.

+---------------------------+-----------------------------------------------------+
| Product Name              | Description                                         |
+===========================+=====================================================+
| i01                       | I01 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| i02                       | I02 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| i03                       | I03 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| i04                       | I04 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| i05                       | I05 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| i01_rad                   | I01 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| i02_rad                   | I02 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| i03_rad                   | I03 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| i04_rad                   | I04 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| i05_rad                   | I05 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m01                       | M01 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m02                       | M02 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m03                       | M03 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m04                       | M04 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m05                       | M05 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m06                       | M06 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m07                       | M07 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m08                       | M08 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m09                       | M09 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m10                       | M10 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m11                       | M11 Reflectance Band                                |
+---------------------------+-----------------------------------------------------+
| m12                       | M12 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| m13                       | M13 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| m14                       | M14 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| m15                       | M15 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| m16                       | M16 Brightness Temperature Band                     |
+---------------------------+-----------------------------------------------------+
| m01_rad                   | M01 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m02_rad                   | M02 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m03_rad                   | M03 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m04_rad                   | M04 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m05_rad                   | M05 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m06_rad                   | M06 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m07_rad                   | M07 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m08_rad                   | M08 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m09_rad                   | M09 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m10_rad                   | M10 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m11_rad                   | M11 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m12_rad                   | M12 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m13_rad                   | M13 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m14_rad                   | M14 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m15_rad                   | M15 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| m16_rad                   | M16 Radiance Band                                   |
+---------------------------+-----------------------------------------------------+
| dnb                       | Raw DNB Band (not useful for images)                |
+---------------------------+-----------------------------------------------------+
| histogram_dnb             | Histogram Equalized DNB Band                        |
+---------------------------+-----------------------------------------------------+
| adaptive_dnb              | Adaptive Histogram Equalized DNB Band               |
+---------------------------+-----------------------------------------------------+
| dynamic_dnb               | Dynamic DNB Band from Steve Miller and              |
|                           | Curtis Seaman. Uses erf to scale the data.          |
+---------------------------+-----------------------------------------------------+
| hncc_dnb                  | Simplified High and Near-Constant Contrast          |
|                           | Approach from Stephan Zinke                         |
+---------------------------+-----------------------------------------------------+
| ifog                      | Temperature difference between I05 and I04          |
+---------------------------+-----------------------------------------------------+
| i_solar_zenith_angle      | I Band Solar Zenith Angle                           |
+---------------------------+-----------------------------------------------------+
| i_solar_azimuth_angle     | I Band Solar Azimuth Angle                          |
+---------------------------+-----------------------------------------------------+
| i_sat_zenith_angle        | I Band Satellite Zenith Angle                       |
+---------------------------+-----------------------------------------------------+
| i_sat_azimuth_angle       | I Band Satellite Azimuth Angle                      |
+---------------------------+-----------------------------------------------------+
| m_solar_zenith_angle      | M Band Solar Zenith Angle                           |
+---------------------------+-----------------------------------------------------+
| m_solar_azimuth_angle     | M Band Solar Azimuth Angle                          |
+---------------------------+-----------------------------------------------------+
| m_sat_zenith_angle        | M Band Satellite Zenith Angle                       |
+---------------------------+-----------------------------------------------------+
| m_sat_azimuth_angle       | M Band Satellite Azimuth Angle                      |
+---------------------------+-----------------------------------------------------+
| dnb_solar_zenith_angle    | DNB Band Solar Zenith Angle                         |
+---------------------------+-----------------------------------------------------+
| dnb_solar_azimuth_angle   | DNB Band Solar Azimuth Angle                        |
+---------------------------+-----------------------------------------------------+
| dnb_sat_zenith_angle      | DNB Band Satellite Zenith Angle                     |
+---------------------------+-----------------------------------------------------+
| dnb_sat_azimuth_angle     | DNB Band Satellite Azimuth Angle                    |
+---------------------------+-----------------------------------------------------+
| dnb_lunar_zenith_angle    | DNB Band Lunar Zenith Angle                         |
+---------------------------+-----------------------------------------------------+
| dnb_lunar_azimuth_angle   | DNB Band Lunar Azimuth Angle                        |
+---------------------------+-----------------------------------------------------+
| true_color                | Ratio sharpened rayleigh corrected true color       |
+---------------------------+-----------------------------------------------------+
| false_color               | Ratio sharpened rayleigh corrected false color      |
+---------------------------+-----------------------------------------------------+

"""

from __future__ import annotations

from argparse import ArgumentParser, _ArgumentGroup
from typing import Optional

from satpy import DataQuery, Scene

from polar2grid.core.script_utils import ExtendConstAction

from ._base import ReaderProxyBase

PREFERRED_CHUNK_SIZE: int = 6400

I_PRODUCTS = [
    "I01",
    "I02",
    "I03",
    "I04",
    "I05",
]
I_ANGLE_PRODUCTS = [
    "i_solar_zenith_angle",
    "i_solar_azimuth_angle",
    "i_sat_zenith_angle",
    "i_sat_azimuth_angle",
]
M_PRODUCTS = [
    "M01",
    "M02",
    "M03",
    "M04",
    "M05",
    "M06",
    "M07",
    "M08",
    "M09",
    "M10",
    "M11",
    "M12",
    "M13",
    "M14",
    "M15",
    "M16",
]
M_ANGLE_PRODUCTS = [
    "m_solar_zenith_angle",
    "m_solar_azimuth_angle",
    "m_sat_zenith_angle",
    "m_sat_azimuth_angle",
]
DNB_PRODUCTS = [
    "histogram_dnb",
    "adaptive_dnb",
    "dynamic_dnb",
    "hncc_dnb",
]
DNB_ANGLE_PRODUCTS = [
    "dnb_solar_zenith_angle",
    "dnb_solar_azimuth_angle",
    "dnb_sat_zenith_angle",
    "dnb_sat_azimuth_angle",
    "dnb_lunar_zenith_angle",
    "dnb_lunar_azimuth_angle",
]
TRUE_COLOR_PRODUCTS = ["true_color"]
FALSE_COLOR_PRODUCTS = ["false_color"]
OTHER_COMPS = [
    "ifog",
]


PRODUCT_ALIASES = {}


def _process_legacy_and_rad_products(satpy_names, band_aliases, rad_aliases):
    """Map all lowercase band names to uppercase names and add radiance product."""
    for band in satpy_names:
        # P2G name is lowercase, Satpy is uppercase
        PRODUCT_ALIASES[band.lower()] = band
        band_aliases.append(band.lower())

        # radiance products for M and I bands
        rad_name = band.lower() + "_rad"
        dq = DataQuery(name=band, calibration="radiance")
        PRODUCT_ALIASES[rad_name] = dq
        rad_aliases.append(rad_name)


I_ALIASES = []
I_RAD_PRODUCTS = []
_process_legacy_and_rad_products(I_PRODUCTS, I_ALIASES, I_RAD_PRODUCTS)
M_ALIASES = []
M_RAD_PRODUCTS = []
_process_legacy_and_rad_products(M_PRODUCTS, M_ALIASES, M_RAD_PRODUCTS)

_AWIPS_TRUE_COLOR = ["viirs_crefl08", "viirs_crefl04", "viirs_crefl03"]
_AWIPS_FALSE_COLOR = ["viirs_crefl07", "viirs_crefl09", "viirs_crefl08"]

PRODUCT_ALIASES["dnb_solar_zenith_angle"] = DataQuery(name="dnb_solar_zenith_angle")
PRODUCT_ALIASES["dnb_solar_azimuth_angle"] = DataQuery(name="dnb_solar_azimuth_angle")
PRODUCT_ALIASES["dnb_sat_zenith_angle"] = DataQuery(name="dnb_satellite_zenith_angle")
PRODUCT_ALIASES["dnb_sat_azimuth_angle"] = DataQuery(name="dnb_satellite_azimuth_angle")
PRODUCT_ALIASES["dnb_lunar_zenith_angle"] = DataQuery(name="dnb_lunar_zenith_angle")
PRODUCT_ALIASES["dnb_lunar_azimuth_angle"] = DataQuery(name="dnb_lunar_azimuth_angle")
PRODUCT_ALIASES["m_solar_zenith_angle"] = DataQuery(name="solar_zenith_angle", resolution=742)
PRODUCT_ALIASES["m_solar_azimuth_angle"] = DataQuery(name="solar_azimuth_angle", resolution=742)
PRODUCT_ALIASES["m_sat_zenith_angle"] = DataQuery(name="satellite_zenith_angle", resolution=742)
PRODUCT_ALIASES["m_sat_azimuth_angle"] = DataQuery(name="satellite_azimuth_angle", resolution=742)
PRODUCT_ALIASES["i_solar_zenith_angle"] = DataQuery(name="solar_zenith_angle", resolution=371)
PRODUCT_ALIASES["i_solar_azimuth_angle"] = DataQuery(name="solar_azimuth_angle", resolution=371)
PRODUCT_ALIASES["i_sat_zenith_angle"] = DataQuery(name="satellite_zenith_angle", resolution=371)
PRODUCT_ALIASES["i_sat_azimuth_angle"] = DataQuery(name="satellite_azimuth_angle", resolution=371)

DEFAULT_PRODUCTS = I_ALIASES + M_ALIASES + DNB_PRODUCTS[1:] + TRUE_COLOR_PRODUCTS + FALSE_COLOR_PRODUCTS + OTHER_COMPS
P2G_PRODUCTS = I_ALIASES + M_ALIASES + DNB_PRODUCTS + I_RAD_PRODUCTS + M_RAD_PRODUCTS
P2G_PRODUCTS += I_ANGLE_PRODUCTS + M_ANGLE_PRODUCTS + DNB_ANGLE_PRODUCTS + OTHER_COMPS
P2G_PRODUCTS += TRUE_COLOR_PRODUCTS + FALSE_COLOR_PRODUCTS
P2G_PRODUCTS += [
    "viirs_crefl01",
    "viirs_crefl02",
    "viirs_crefl03",
    "viirs_crefl04",
    "viirs_crefl07",
    "viirs_crefl08",
    "viirs_crefl09",
]
P2G_PRODUCTS += ["dynamic_dnb_saturation", "unsharp_true_color"]

FILTERS = {
    "day_only": {
        "standard_name": [
            "toa_bidirectional_reflectance",
            "true_color",
            "false_color",
            "natural_color",
            "corrected_reflectance",
        ],
    },
    "night_only": {
        "standard_name": ["temperature_difference"],
    },
}


class ReaderProxy(ReaderProxyBase):
    """Provide Polar2Grid-specific information about this reader's products."""

    is_polar2grid_reader = True

    def __init__(self, scn: Scene, user_products: list[str]):
        self._modified_aliases = PRODUCT_ALIASES.copy()
        if "dynamic_dnb_saturation" in user_products:
            # they specified --dnb-saturation-correction
            # let's modify the aliases so dynamic_dnb points to this product
            user_products.remove("dynamic_dnb_saturation")
            self._modified_aliases["dynamic_dnb"] = DataQuery(name="dynamic_dnb_saturation")
        super().__init__(scn, user_products)

    def get_default_products(self) -> list[str]:
        """Get products to load if users hasn't specified any others."""
        return DEFAULT_PRODUCTS

    def get_all_products(self):
        """Get all polar2grid products that could be loaded."""
        return P2G_PRODUCTS

    @property
    def _aliases(self):
        return self._modified_aliases


def add_reader_argument_groups(
    parser: ArgumentParser, group: Optional[_ArgumentGroup] = None
) -> tuple[Optional[_ArgumentGroup], Optional[_ArgumentGroup]]:
    """Add reader-specific command line arguments to an existing argument parser.

    If ``group`` is provided then arguments are added to this group. If not,
    a new group is added to the parser and arguments added to this new group.

    """
    if group is None:
        group = parser.add_argument_group(title="VIIRS SDR Reader")

    group.add_argument(
        "--i-bands",
        dest="products",
        action=ExtendConstAction,
        const=I_ALIASES,
        help="Add all I-band raw products to list of products",
    )
    group.add_argument(
        "--m-bands",
        dest="products",
        action=ExtendConstAction,
        const=M_ALIASES,
        help="Add all M-band raw products to list of products",
    )
    group.add_argument(
        "--dnb-angle-products",
        dest="products",
        action=ExtendConstAction,
        const=DNB_ANGLE_PRODUCTS,
        help="Add DNB-band geolocation 'angle' products to list of products",
    )
    group.add_argument(
        "--dnb-saturation-correction",
        dest="products",
        action=ExtendConstAction,
        const=["dynamic_dnb_saturation"],
        help="Enable dynamic DNB saturation correction (normally used for aurora scenes)",
    )
    group.add_argument(
        "--i-angle-products",
        dest="products",
        action=ExtendConstAction,
        const=I_ANGLE_PRODUCTS,
        help="Add I-band geolocation 'angle' products to list of products",
    )
    group.add_argument(
        "--m-angle-products",
        dest="products",
        action=ExtendConstAction,
        const=M_ANGLE_PRODUCTS,
        help="Add M-band geolocation 'angle' products to list of products",
    )
    group.add_argument(
        "--m-rad-products",
        dest="products",
        action=ExtendConstAction,
        const=M_RAD_PRODUCTS,
        help="Add M-band geolocation radiance products to list of products",
    )
    group.add_argument(
        "--i-rad-products",
        dest="products",
        action=ExtendConstAction,
        const=I_RAD_PRODUCTS,
        help="Add I-band geolocation radiance products to list of products",
    )
    group.add_argument(
        "--awips-true-color",
        dest="products",
        action=ExtendConstAction,
        const=_AWIPS_TRUE_COLOR,
        help="Add individual CREFL corrected products to create the 'true_color' composite in AWIPS.",
    )
    group.add_argument(
        "--awips-false-color",
        dest="products",
        action=ExtendConstAction,
        const=_AWIPS_FALSE_COLOR,
        help="Add individual CREFL corrected products to create the 'false_color' composite in AWIPS.",
    )
    return group, None



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2022 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Tests for filtering utilities."""

import dask
import dask.array as da
import numpy as np
import pytest
from pyresample import SwathDefinition
from pyresample.boundary import AreaBoundary, Boundary
from pytest_lazyfixture import lazy_fixture
from satpy.tests.utils import CustomScheduler

from polar2grid.filters._utils import boundary_for_area

from .._fixture_utils import generate_lonlat_data


def _swath_def_nan_rows() -> SwathDefinition:
    lons, lats = generate_lonlat_data((200, 100))
    lons[:9, :] = np.nan
    lats[:9, :] = np.nan
    lons_da = da.from_array(lons)
    lats_da = da.from_array(lats)
    swath_def = SwathDefinition(
        lons_da,
        lats_da,
    )
    return swath_def


def _swath_def_antimeridian_nan_rows() -> SwathDefinition:
    lons, lats = generate_lonlat_data((200, 100))
    lons = lons - 115.0
    lons[lons <= -180] += 360
    lons[:9, :] = np.nan
    lats[:9, :] = np.nan
    lons_da = da.from_array(lons)
    lats_da = da.from_array(lats)
    swath_def = SwathDefinition(
        lons_da,
        lats_da,
    )
    return swath_def


def _swath_def_nans_in_right_column() -> SwathDefinition:
    lons, lats = generate_lonlat_data((200, 100))
    lons[:100, -1] = np.nan
    lons_da = da.from_array(lons)
    lats_da = da.from_array(lats)
    swath_def = SwathDefinition(
        lons_da,
        lats_da,
    )
    return swath_def


def _exp_boundary_nan_rows():
    exp_boundary = AreaBoundary(
        ([-71.5, -40.907036], [60.5, 60.5]),
        ([-40.907036, -40.907036], [60.5, 23.721106]),
        ([-40.907036, -71.5], [23.721106, 23.721106]),
        ([-71.5, -40.907036], [23.721106, 60.5]),
    )
    return exp_boundary


def _exp_boundary_antimeridian_nan_rows():
    exp_boundary = AreaBoundary(
        ([-71.5 - 115.0 + 360.0, -40.907036 - 115.0], [60.5, 60.5]),
        ([-40.907036 - 115.0, -40.907036 - 115.0], [60.5, 23.721106]),
        ([-40.907036 - 115.0, -71.5 - 115.0 + 360.0], [23.721106, 23.721106]),
        ([-71.5 - 115.0 + 360.0, -40.907036 - 115.0 + 360.0], [23.721106, 60.5]),
    )
    return exp_boundary


def _exp_boundary_nans_in_right_column():
    return Boundary([-40.5, -40.907036], [22.5, 23.721106])


def _exp_boundary_geos_area():
    lons = np.array([-113.08581059, -113.08581059])
    lats = np.array([15.11098716, 15.11098716])
    exp_boundary = Boundary(lons, lats)
    return exp_boundary


@pytest.mark.parametrize(
    ("geom_or_func", "exp_func"),
    [
        (_swath_def_nan_rows, _exp_boundary_nan_rows),
        (_swath_def_antimeridian_nan_rows, _exp_boundary_antimeridian_nan_rows),
        (_swath_def_nans_in_right_column, _exp_boundary_nans_in_right_column),
        (lazy_fixture("goes_east_conus_area_def"), _exp_boundary_geos_area),
    ],
)
def test_boundary_for_area(geom_or_func, exp_func):
    geom_obj = geom_or_func() if callable(geom_or_func) else geom_or_func
    exp_boundary = exp_func()

    with dask.config.set(scheduler=CustomScheduler(1)):
        boundary = boundary_for_area(geom_obj)

    np.testing.assert_allclose(boundary.contour()[0][0], exp_boundary.contour()[0][0])
    np.testing.assert_allclose(boundary.contour()[0][-1], exp_boundary.contour()[0][-1])
    np.testing.assert_allclose(boundary.contour()[1][0], exp_boundary.contour()[1][0])
    np.testing.assert_allclose(boundary.contour()[1][-1], exp_boundary.contour()[1][-1])



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2014-2021 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
#     input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Classes for metadata operations in polar2grid."""

import logging

import numpy
from pyproj import Proj

LOG = logging.getLogger(__name__)


class GridDefinition(dict):
    """Projected grid defined by a PROJ.4 projection string and other grid parameters.

    Required Information:
        - grid_name: Identifying name for the grid
        - proj4_definition (string): PROJ.4 projection definition
        - height: Height of the grid in number of pixels
        - width: Width of the grid in number of pixels
        - cell_height: Grid cell height in the projection domain (usually in meters or degrees)
        - cell_width: Grid cell width in the projection domain (usually in meters or degrees)
        - origin_x: X-coordinate of the upper-left corner of the grid in the projection domain
        - origin_y: Y-coordinate of the upper-left corner of the grid in the projection domain
    """

    required_kwargs = (
        "grid_name",
        "proj4_definition",
        "height",
        "width",
        "cell_height",
        "cell_width",
        "origin_x",
        "origin_y",
    )

    def __init__(self, *args, **kwargs):
        # pyproj.Proj object if needed
        self.p = None
        if "proj4_definition" in kwargs:
            # pyproj doesn't like unicode
            kwargs["proj4_definition"] = str(kwargs["proj4_definition"])
        super(GridDefinition, self).__init__(*args, **kwargs)
        self.validate_keys()

    def validate_keys(self):
        # sanity check, does this dictionary have everything the class expects it to
        for k in self.required_kwargs:
            if k not in self:
                raise ValueError("Missing required keyword '%s'" % (k,))

    def __str__(self):
        keys = sorted(self.keys())
        keys.insert(0, keys.pop(keys.index("height")))
        keys.insert(0, keys.pop(keys.index("width")))
        keys.insert(0, keys.pop(keys.index("cell_height")))
        keys.insert(0, keys.pop(keys.index("cell_width")))
        keys.insert(0, keys.pop(keys.index("proj4_definition")))
        keys.insert(0, keys.pop(keys.index("grid_name")))
        return "\n".join("%s: %s" % (k, self[k]) for k in keys)

    @property
    def proj(self):
        if self.p is None:
            self.p = Proj(self["proj4_definition"])
        return self.p

    @property
    def is_static(self):
        return all(
            [self[x] is not None for x in ["height", "width", "cell_height", "cell_width", "origin_x", "origin_y"]]
        )

    @property
    def lonlat_lowerleft(self):
        x_ll, y_ll = self.xy_lowerleft
        return self.proj(x_ll, y_ll, inverse=True)

    @property
    def lonlat_lowerright(self):
        x_ll, y_ll = self.xy_lowerright
        return self.proj(x_ll, y_ll, inverse=True)

    @property
    def lonlat_upperright(self):
        x_ur, y_ur = self.xy_upperright
        return self.proj(x_ur, y_ur, inverse=True)

    @property
    def lonlat_upperleft(self):
        x_ur, y_ur = self["origin_x"], self["origin_y"]
        return self.proj(x_ur, y_ur, inverse=True)

    @property
    def lonlat_center(self):
        x_pixel = self["width"] / 2
        y_pixel = self["height"] / 2
        return self.get_lonlat(x_pixel, y_pixel)

    def get_lonlat(self, x_pixel, y_pixel):
        x = self["origin_x"] + x_pixel * self["cell_width"]
        y = self["origin_y"] + y_pixel * self["cell_height"]
        return self.proj(x, y, inverse=True)

    @property
    def xy_lowerright(self):
        y_ll = self["origin_y"] + self["cell_height"] * (self["height"] - 1)
        x_ll = self["origin_x"] + self["cell_width"] * (self["width"] - 1)
        return x_ll, y_ll

    @property
    def xy_lowerleft(self):
        y_ll = self["origin_y"] + self["cell_height"] * (self["height"] - 1)
        return self["origin_x"], y_ll

    @property
    def xy_upperright(self):
        x_ll = self["origin_x"] + self["cell_width"] * (self["width"] - 1)
        return x_ll, self["origin_y"]

    @property
    def proj4_dict(self):
        parts = [x.replace("+", "") for x in self["proj4_definition"].split(" ")]
        no_defs = False
        if "no_defs" in parts:
            parts.remove("no_defs")
            no_defs = True
        over = False
        if "over" in parts:
            parts.remove("over")
            over = True

        proj4_dict = dict(p.split("=") for p in parts)
        # Convert numeric parameters to floats
        for k in [
            "lat_0",
            "lat_1",
            "lat_2",
            "lat_ts",
            "lat_b",
            "lat_t",
            "lon_0",
            "lon_1",
            "lon_2",
            "lonc",
            "a",
            "b",
            "f",
            "es",
            "h",
        ]:
            if k in proj4_dict:
                proj4_dict[k] = float(proj4_dict[k])

        # load information from PROJ.4 about the ellipsis if possible
        if "ellps" in proj4_dict and ("a" not in proj4_dict or "b" not in proj4_dict):
            import pyproj

            ellps = pyproj.pj_ellps[proj4_dict["ellps"]]
            proj4_dict["a"] = ellps["a"]
            if "b" not in ellps and "rf" in ellps:
                proj4_dict["f"] = 1.0 / ellps["rf"]
            else:
                proj4_dict["b"] = ellps["b"]

        if "a" in proj4_dict and "f" in proj4_dict and "b" not in proj4_dict:
            # add a "b" attribute back in if they used "f" instead
            proj4_dict["b"] = proj4_dict["a"] * (1 - proj4_dict["f"])

        # Add removed keywords back in
        if no_defs:
            proj4_dict["no_defs"] = True
        if over:
            proj4_dict["over"] = True

        return proj4_dict

    @property
    def gdal_geotransform(self):
        # GDAL PixelIsArea (default) geotiffs expects coordinates for the
        # upper-left corner of the pixel
        # Polar2Grid and satellite imagery formats assume coordinates for
        # center of the pixel
        half_pixel_x = self["cell_width"] / 2.0
        half_pixel_y = self["cell_height"] / 2.0
        # cell_height is negative so we need to substract instead of add
        return (
            self["origin_x"] - half_pixel_x,
            self["cell_width"],
            0,
            self["origin_y"] - half_pixel_y,
            0,
            self["cell_height"],
        )

    def get_xy_arrays(self):
        # the [None] indexing adds a dimension to the array
        x = self["origin_x"] + numpy.repeat(
            numpy.arange(self["width"])[None] * self["cell_width"], self["height"], axis=0
        )
        y = self["origin_y"] + numpy.repeat(
            numpy.arange(self["height"])[:, None] * self["cell_height"], self["width"], axis=1
        )
        return x, y

    def get_geolocation_arrays(self):
        """Calculate longitude and latitude arrays for the grid.

        :returns: (longitude array, latitude array)
        """
        x, y = self.get_xy_arrays()
        return self.proj(x, y, inverse=True)

    def to_basemap_object(self):
        from mpl_toolkits.basemap import Basemap

        proj4_dict = self.proj4_dict
        proj4_dict.pop("no_defs", None)
        proj4_dict.pop("units", None)
        proj4_dict["projection"] = proj4_dict.pop("proj")

        if "a" in proj4_dict and "b" in proj4_dict:
            a = proj4_dict.pop("a")
            b = proj4_dict.pop("b")
            proj4_dict["rsphere"] = (a, b)
        elif "a" in proj4_dict:
            proj4_dict["rsphere"] = proj4_dict.pop("a")
        elif "b" in proj4_dict:
            proj4_dict["rsphere"] = proj4_dict.pop("b")

        lon_ll, lat_ll = self.lonlat_lowerleft
        lon_ur, lat_ur = self.lonlat_upperright
        LOG.debug("Passing basemap the following keywords from PROJ.4: %r", proj4_dict)
        LOG.debug("Lower-left corner: (%f, %f); Upper-right corner: (%f, %f)", lon_ll, lat_ll, lon_ur, lat_ur)
        return Basemap(llcrnrlon=lon_ll, llcrnrlat=lat_ll, urcrnrlon=lon_ur, urcrnrlat=lat_ur, **proj4_dict)

    @property
    def ll_extent(self):
        xy_ll = self.xy_lowerleft
        # NOTE: cell_height is negative
        return xy_ll[0] - self["cell_width"] / 2.0, xy_ll[1] + self["cell_height"] / 2.0

    @property
    def ll_extent_lonlat(self):
        return self.proj(*self.ll_extent, inverse=True)

    @property
    def ur_extent(self):
        xy_ur = self.xy_upperright
        # NOTE: cell_height is negative
        return xy_ur[0] + self["cell_width"] / 2.0, xy_ur[1] - self["cell_height"] / 2.0

    @property
    def ur_extent_lonlat(self):
        return self.proj(*self.ur_extent, inverse=True)

    def to_satpy_area(self):
        from pyresample.geometry import AreaDefinition, DynamicAreaDefinition

        if self.is_static:
            xy_ll = self.ll_extent
            xy_ur = self.ur_extent
            return AreaDefinition(
                self["grid_name"],
                self["grid_name"],
                self["grid_name"],
                self.proj4_dict,
                self["width"],
                self["height"],
                area_extent=xy_ll + xy_ur,
            )
        kwargs = {}
        if self["cell_width"] is not None:
            kwargs["resolution"] = (abs(self["cell_width"]), abs(self["cell_height"]))
        return DynamicAreaDefinition(
            self["grid_name"], self["grid_name"], self.proj4_dict, self["width"], self["height"], **kwargs
        )



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2019 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
#
#     Written by William Roberts and David Hoese    May 2019
#     University of Wisconsin-Madison
#     Space Science and Engineering Center
#     1225 West Dayton Street
#     Madison, WI  53706
#     wroberts4@wisc.edu and david.hoese@ssec.wisc.edu
"""The VIRR Level 1B reader operates on Level 1B (L1B) HDF5 files.

Files usually have the following naming scheme:

    tf2018343030324.FY3C-L_VIRRX_L1B.HDF or tf2018343092538.FY3B-L_VIRRX_L1B.HDF
    the numbers at the start are year julian-day hour minute seconds

This reader's default resampling algorithm is ``ewa`` for Elliptical Weighted
Averaging resampling. The ``--fornav-D`` parameter is set to 40 and the
``--fornav-d`` parameter is set to 1.

+---------------------------+-----------------------------------------------------+-------------------------+
| **Product Name**          | **Description**                                     | Central Wavelength (um) |
+===========================+=====================================================+=========================+
| 1                         | Channel 1 Reflectance Band                          |0.63                     |
+---------------------------+-----------------------------------------------------+-------------------------+
| 2                         | Channel 2 Reflectance Band                          |0.865                    |
+---------------------------+-----------------------------------------------------+-------------------------+
| 3                         | Channel 3 Emissive Band                             |3.74                     |
+---------------------------+-----------------------------------------------------+-------------------------+
| 4                         | Channel 4 Emissive Band                             |10.8                     |
+---------------------------+-----------------------------------------------------+-------------------------+
| 5                         | Channel 5 Emissive Band                             |12.0                     |
+---------------------------+-----------------------------------------------------+-------------------------+
| 6                         | Channel 6 Reflectance Band                          |1.60                     |
+---------------------------+-----------------------------------------------------+-------------------------+
| 7                         | Channel 7 Reflectance Band                          |0.455                    |
+---------------------------+-----------------------------------------------------+-------------------------+
| 8                         | Channel 8 Reflectance Band                          |0.505                    |
+---------------------------+-----------------------------------------------------+-------------------------+
| 9                         | Channel 9 Reflectance Band                          |0.555                    |
+---------------------------+-----------------------------------------------------+-------------------------+
| 10                        | Channel 10 Reflectance Band                         |1.36                     |
+---------------------------+-----------------------------------------------------+-------------------------+
| true_color                | Ratio sharpened rayleigh corrected true color       |N/A                      |
+---------------------------+-----------------------------------------------------+-------------------------+
"""
from __future__ import annotations

from argparse import ArgumentParser, _ArgumentGroup
from typing import Optional

from ._base import ReaderProxyBase

ALL_BANDS = [str(x) for x in range(1, 11)]
ALL_COMPS = ["true_color"]
ALL_ANGLES = ["solar_zenith_angle", "solar_azimuth_angle", "sensor_zenith_angle", "sensor_azimuth_angle"]

DEFAULT_PRODUCTS = ALL_BANDS + ALL_ANGLES + ALL_COMPS

PRODUCT_ALIASES = {}

FILTERS = {
    "day_only": {
        "standard_name": [
            "toa_bidirectional_reflectance",
            "true_color",
            "corrected_reflectance",
        ],
    },
}


class ReaderProxy(ReaderProxyBase):
    """Provide Polar2Grid-specific information about this reader's products."""

    is_polar2grid_reader = True

    def get_default_products(self) -> list[str]:
        """Get products to load if users hasn't specified any others."""
        return DEFAULT_PRODUCTS

    def get_all_products(self) -> list[str]:
        """Get all polar2grid products that could be loaded."""
        return ALL_BANDS + ALL_ANGLES + ALL_COMPS

    @property
    def _aliases(self) -> dict:
        return PRODUCT_ALIASES


def add_reader_argument_groups(
    parser: ArgumentParser, group: Optional[_ArgumentGroup] = None
) -> tuple[Optional[_ArgumentGroup], Optional[_ArgumentGroup]]:
    """Add reader-specific command line arguments to an existing argument parser."""
    if group is None:
        group = parser.add_argument_group(title="VIRR L1b Reader")
    return group, None



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2022 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Shared utilities between fixtures."""
from __future__ import annotations

from datetime import datetime

import numpy as np
from numpy.typing import DTypeLike, NDArray
from satpy import DatasetDict, Scene
from satpy.readers.yaml_reader import FileYAMLReader

START_TIME = datetime(2021, 1, 1, 12, 0, 0)


class _FakeReader(FileYAMLReader):
    """Fake reader to avoid loading real files during testing."""

    def __init__(self, reader_name, dataset_dict, all_dataset_ids, available_dataset_ids):
        self._name = reader_name
        self._forced_all_ids = all_dataset_ids
        self._forced_available_ids = available_dataset_ids
        self._forced_products = DatasetDict(dataset_dict)

    @property
    def start_time(self):
        return START_TIME

    @property
    def end_time(self):
        return START_TIME

    @property
    def sensor_names(self):
        return {
            "viirs_sdr": {"viirs"},
            "abi_l1b": {"abi"},
            "avhrr_l1b_aapp": {"avhrr-3"},
        }[self._name]

    @property
    def all_dataset_ids(self):
        return self._forced_all_ids

    @property
    def available_dataset_ids(self):
        return self._forced_available_ids

    def load(self, dataset_keys, previous_datasets=None, **kwargs):
        new_result = DatasetDict()
        for data_query in dataset_keys:
            try:
                data_arr = self._forced_products[data_query]
            except KeyError:
                continue
            data_arr.attrs.update(data_query.to_dict())
            new_result[data_query] = data_arr
        return new_result


class _TestingScene(Scene):
    """Special Scene class to mimic a real Scene that would be created during normal execution."""

    def __init__(self, *args, data_array_dict=None, all_dataset_ids=None, available_dataset_ids=None, **kwargs):
        self._forced_all_ids = all_dataset_ids
        self._forced_available_ids = available_dataset_ids
        self._forced_products = data_array_dict

        super().__init__(*args, **kwargs)

    def _create_reader_instances(self, filenames=None, reader=None, reader_kwargs=None):
        if reader is None or filenames is None or self._forced_all_ids is None or self._forced_available_ids is None:
            return {}
        return {
            reader: _FakeReader(reader, self._forced_products, self._forced_all_ids, self._forced_available_ids),
        }


def generate_lonlat_data(shape: tuple[int, int], dtype: DTypeLike = np.float32) -> tuple[NDArray, NDArray]:
    lat = np.repeat(np.linspace(25.0, 55.0, shape[0])[:, None], shape[1], 1)
    lat *= np.linspace(0.9, 1.1, shape[1])
    lon = np.repeat(np.linspace(-45.0, -65.0, shape[1])[None, :], shape[0], 0)
    lon *= np.linspace(0.9, 1.1, shape[0])[:, None]
    return lon.astype(dtype), lat.astype(dtype)



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2016-2021 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
#     input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
#
# Written by David Hoese    July 2016
# University of Wisconsin-Madison
# Space Science and Engineering Center
# 1225 West Dayton Street
# Madison, WI  53706
# david.hoese@ssec.wisc.edu
"""Script to add coastlines and borders to a geotiff while also creating a PNG."""
from __future__ import annotations

import argparse
import logging
import os
import sys

import numpy as np
import rasterio
from aggdraw import Font
from PIL import Image, ImageFont
from pkg_resources import resource_filename as get_resource_filename
from pycoast import ContourWriterAGG
from pydecorate import DecoratorAGG
from pyresample.utils import get_area_def_from_raster
from trollimage.colormap import Colormap

from polar2grid.utils.config import add_polar2grid_config_paths

LOG = logging.getLogger(__name__)
PYCOAST_DIR = os.environ.get("GSHHS_DATA_ROOT")


def _get_rio_colormap(rio_ds, bidx):
    try:
        return rio_ds.colormap(bidx)
    except ValueError:
        return None


def _get_colorbar_vmin_vmax(arg_min, arg_max, metadata, input_dtype):
    scale = metadata.get("scale", metadata.get("scale_factor"))
    offset = metadata.get("offset", metadata.get("add_offset"))
    dtype_min = float(np.iinfo(input_dtype).min)
    dtype_max = float(np.iinfo(input_dtype).max)
    if arg_min is None and scale is None:
        LOG.warning(
            "Colorbar min/max metadata not found and not provided "
            "on the command line. Defaulting to data type limits."
        )
        return dtype_min, dtype_max

    if arg_min is not None:
        vmin = float(arg_min)
        vmax = float(arg_max)
    else:
        scale, offset = _convert_scale_offset_to_float(scale, offset)
        vmin, vmax = _vmin_vmax_from_scale_offset(scale, offset, dtype_min, dtype_max)
    return vmin, vmax


def _convert_scale_offset_to_float(scale: str, offset: str) -> tuple[float, float]:
    scale = float(scale)
    offset = float(offset)
    if np.isnan(scale) or np.isnan(offset):
        raise ValueError(
            "Can't automatically set colorbar limits with "
            "geotiff metadata as scale/offset are set to "
            "NaN. This indicates a non-linear enhancement or "
            "RGB/A composite that was enhanced. These cases "
            "cannot be represented properly by a colorbar."
        )
    return scale, offset


def _vmin_vmax_from_scale_offset(scale: float, offset: float, dtype_min: int, dtype_max: int) -> tuple[float, float]:
    delta = dtype_max - dtype_min
    vmin = offset
    vmax = delta * scale + offset
    # floating point error made it not an integer
    if abs(vmin - np.round(vmin, 0)) <= 0.001:
        vmin = np.round(vmin, 0)
    if abs(vmax - np.round(vmax, 0)) <= 0.001:
        vmax = np.round(vmax, 0)
    return vmin, vmax


def _apply_decorator_alignment(dc, align):
    if align == "top":
        dc.align_top()
        dc.write_horizontally()
    elif align == "bottom":
        dc.align_bottom()
        dc.write_horizontally()
    elif align == "left":
        dc.align_left()
        dc.write_vertically()
    elif align == "right":
        dc.align_right()
        dc.write_vertically()


def _add_colorbar_to_image(img, font, tick_color, align, **kwargs):
    dc = DecoratorAGG(img)
    _apply_decorator_alignment(dc, align)
    dc.add_scale(
        font=font,
        line=tick_color,
        **kwargs,
    )


def get_parser():
    parser = argparse.ArgumentParser(
        description="Add overlays to a GeoTIFF file and save as a PNG file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    _add_coastlines_arguments(parser)
    _add_rivers_arguments(parser)
    _add_grid_arguments(parser)
    _add_borders_arguments(parser)
    _add_colorbar_arguments(parser)
    _add_global_arguments(parser)
    parser.add_argument("input_tiff", nargs="+", help="Input geotiff(s) to process")
    return parser


def _add_coastlines_arguments(parser: argparse.ArgumentParser) -> None:
    group = parser.add_argument_group("coastlines")
    group.add_argument("--add-coastlines", action="store_true", help="Add coastlines")
    group.add_argument(
        "--coastlines-resolution",
        choices="clihf",
        default="i",
        help="Resolution of coastlines to add (crude, low, intermediate, high, full)",
    )
    group.add_argument(
        "--coastlines-level",
        choices=range(1, 7),
        type=int,
        default=4,
        help="Level of detail from the selected resolution dataset",
    )
    group.add_argument(
        "--coastlines-outline",
        default=["yellow"],
        nargs="*",
        help="Color of coastline lines (color name or 3 RGB integers)",
    )
    group.add_argument("--coastlines-fill", default=None, nargs="*", help="Color of land")
    group.add_argument("--coastlines-width", default=1.0, type=float, help="Width of coastline lines")


def _add_rivers_arguments(parser: argparse.ArgumentParser) -> None:
    group = parser.add_argument_group("rivers")
    group.add_argument("--add-rivers", action="store_true", help="Add rivers grid")
    group.add_argument(
        "--rivers-resolution",
        choices="clihf",
        default="c",
        help="Resolution of rivers to add (crude, low, intermediate, high, full)",
    )
    group.add_argument(
        "--rivers-level", choices=range(0, 11), type=int, default=5, help="Level of detail for river lines"
    )
    group.add_argument(
        "--rivers-outline", default=["blue"], nargs="*", help="Color of river lines (color name or 3 RGB integers)"
    )
    group.add_argument("--rivers-width", default=1.0, type=float, help="Width of rivers lines")


def _add_grid_arguments(parser: argparse.ArgumentParser) -> None:
    group = parser.add_argument_group("grid")
    group.add_argument("--add-grid", action="store_true", help="Add lat/lon grid")
    group.add_argument("--grid-no-text", dest="grid_text", action="store_false", help="Add labels to lat/lon grid")
    group.add_argument("--grid-text-size", default=32, type=int, help="Lat/lon grid text font size")
    group.add_argument("--grid-font", default="Vera.ttf", help="Path to TTF font (package provided or custom path)")
    group.add_argument(
        "--grid-fill", nargs="*", default=["cyan"], help="Color of grid text (color name or 3 RGB integers)"
    )
    group.add_argument(
        "--grid-outline", nargs="*", default=["cyan"], help="Color of grid lines (color name or 3 RGB integers)"
    )
    group.add_argument(
        "--grid-minor-outline", nargs="*", default=["cyan"], help="Color of tick lines (color name or 3 RGB integers)"
    )
    group.add_argument(
        "--grid-D", nargs=2, default=(10.0, 10.0), type=float, help="Degrees between grid lines (lon, lat)"
    )
    group.add_argument(
        "--grid-d", nargs=2, default=(2.0, 2.0), type=float, help="Degrees between tick lines (lon, lat)"
    )
    group.add_argument(
        "--grid-lon-placement", choices=["tl", "lr", "lc", "cc"], default="tb", help="Longitude label placement"
    )
    group.add_argument(
        "--grid-lat-placement", choices=["tl", "lr", "lc", "cc"], default="lr", help="Latitude label placement"
    )
    group.add_argument("--grid-width", default=1.0, type=float, help="Width of grid lines")


def _add_borders_arguments(parser: argparse.ArgumentParser) -> None:
    group = parser.add_argument_group("borders")
    group.add_argument("--add-borders", action="store_true", help="Add country and/or region borders")
    group.add_argument(
        "--borders-resolution",
        choices="clihf",
        default="i",
        help="Resolution of borders to add (crude, low, intermediate, high, full)",
    )
    group.add_argument(
        "--borders-level", choices=range(1, 4), default=2, type=int, help="Level of detail for border lines"
    )
    group.add_argument(
        "--borders-outline", default=["white"], nargs="*", help="Color of border lines (color name or 3 RGB integers)"
    )
    group.add_argument("--borders-width", default=1.0, type=float, help="Width of border lines")


def _add_colorbar_arguments(parser: argparse.ArgumentParser) -> None:
    group = parser.add_argument_group("colorbar")
    group.add_argument("--add-colorbar", action="store_true", help="Add colorbar on top of image")
    group.add_argument(
        "--colorbar-colormap-file",
        help=argparse.SUPPRESS,
        # help="Specify the colormap file that was used to "
        # "colorize the provided RGB geotiff. Only used if "
        # "the provided geotiff is RGB/A. Otherwise the "
        # "geotiff is expected to include the colormap as "
        # "a geotiff color table.",
    )
    group.add_argument("--colorbar-width", type=int, help="Number of pixels wide")
    group.add_argument("--colorbar-height", type=int, help="Number of pixels high")
    group.add_argument(
        "--colorbar-extend", action="store_true", help="Extend colorbar to full width/height of the image"
    )
    group.add_argument(
        "--colorbar-tick-marks", type=float, default=5.0, help="Major tick and tick label interval in data units"
    )
    group.add_argument("--colorbar-minor-tick-marks", type=float, default=1.0, help="Minor tick interval in data units")
    group.add_argument("--colorbar-text-size", default=32, type=int, help="Tick label font size")
    group.add_argument(
        "--colorbar-text-color", nargs="*", default=["black"], help="Color of tick text (color name or 3 RGB integers)"
    )
    group.add_argument("--colorbar-font", default="Vera.ttf", help="Path to TTF font (package provided or custom path)")
    group.add_argument(
        "--colorbar-align",
        choices=["left", "top", "right", "bottom"],
        default="bottom",
        help="Which side of the image to place the colorbar",
    )
    group.add_argument("--colorbar-vertical", action="store_true", help="DEPRECATED")
    group.add_argument(
        "--colorbar-min",
        type=float,
        help="Minimum data value of the colorbar."
        " Defaults to 'min_in' of input metadata or"
        " minimum value of the data otherwise.",
    )
    group.add_argument(
        "--colorbar-max",
        type=float,
        help="Maximum data value of the colorbar."
        " Defaults to 'max_in' of input metadata or"
        " maximum value of the data otherwise.",
    )
    group.add_argument("--colorbar-units", help="Units marker to include in the colorbar text")
    group.add_argument("--colorbar-title", help="Title shown with the colorbar")


def _add_global_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--shapes-dir",
        default=PYCOAST_DIR,
        help="Specify alternative directory for coastline shape files (default: GSHSS_DATA_ROOT)",
    )
    parser.add_argument(
        "--cache-dir",
        default=None,
        help="Specify directory where cached coastline output can be stored and accessed in later "
        "executions. The cache will never be cleared by this script. Caching depends on the grid "
        "of the image and the decorations added to the image.",
    )
    parser.add_argument(
        "--cache-regenerate",
        action="store_true",
        help="Force regeneration of any cached overlays. Requires '--cache-dir'.",
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output_filename",
        nargs="+",
        help="Specify the output filename (default replace '.tif' with '.png')",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbosity",
        action="count",
        default=0,
        help="each occurrence increases verbosity 1 level through ERROR-WARNING-INFO-DEBUG (default INFO)",
    )


def main(argv=sys.argv[1:]):
    parser = get_parser()
    args = parser.parse_args(argv)

    levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    logging.basicConfig(level=levels[min(3, args.verbosity)])
    add_polar2grid_config_paths()

    if args.output_filename is None:
        args.output_filename = [x[:-3] + "png" for x in args.input_tiff]
    elif len(args.output_filename) != len(args.input_tiff):
        LOG.error("Output filenames must be equal to number of input tiffs")
        return -1

    if not (args.add_borders or args.add_coastlines or args.add_grid or args.add_rivers or args.add_colorbar):
        LOG.error("Please specify one of the '--add-X' options to modify the image")
        return -1

    if args.cache_dir and not os.path.isdir(args.cache_dir):
        LOG.info(f"Creating cache directory: {args.cache_dir}")
        os.makedirs(args.cache_dir, exist_ok=True)

    # we may be dealing with large images that look like decompression bombs
    # let's turn off the check for the image size in PIL/Pillow
    Image.MAX_IMAGE_PIXELS = None
    # gather all options into a single dictionary that we can pass to pycoast
    pycoast_options = _args_to_pycoast_dict(args)
    colorbar_kwargs = _args_to_colorbar_kwargs(args) if args.add_colorbar else {}
    for input_tiff, output_filename in zip(args.input_tiff, args.output_filename):
        _process_one_image(input_tiff, output_filename, pycoast_options, args.shapes_dir, colorbar_kwargs)
    return 0


def _args_to_pycoast_dict(args):
    opts = {}
    if args.add_coastlines:
        opts["coasts"] = _args_to_coastlines_dict(args)
    if args.add_rivers:
        opts["rivers"] = _args_to_rivers_dict(args)
    if args.add_borders:
        opts["borders"] = _args_to_borders_dict(args)
    if args.add_grid:
        opts["grid"] = _args_to_grid_dict(args)
    if args.cache_dir:
        opts["cache"] = {
            # add "add_coastlines" prefix to cached image name
            "file": os.path.join(args.cache_dir, "add_coastlines"),
            "regenerate": args.cache_regenerate,
        }

    return opts


def _args_to_coastlines_dict(args):
    outline = (
        args.coastlines_outline[0]
        if len(args.coastlines_outline) == 1
        else tuple(int(x) for x in args.coastlines_outline)
    )
    if args.coastlines_fill:
        fill = (
            args.coastlines_fill[0] if len(args.coastlines_fill) == 1 else tuple(int(x) for x in args.coastlines_fill)
        )
    else:
        fill = None
    coasts_dict = {
        "resolution": args.coastlines_resolution,
        "level": args.coastlines_level,
        "width": args.coastlines_width,
        "outline": outline,
        "fill": fill,
    }
    return coasts_dict


def _args_to_rivers_dict(args):
    outline = args.rivers_outline[0] if len(args.rivers_outline) == 1 else tuple(int(x) for x in args.rivers_outline)
    rivers_dict = {
        "resolution": args.rivers_resolution,
        "level": args.rivers_level,
        "width": args.rivers_width,
        "outline": outline,
    }
    return rivers_dict


def _args_to_borders_dict(args):
    outline = args.borders_outline[0] if len(args.borders_outline) == 1 else tuple(int(x) for x in args.borders_outline)
    borders_dict = {
        "resolution": args.borders_resolution,
        "level": args.borders_level,
        "width": args.borders_width,
        "outline": outline,
    }
    return borders_dict


def _args_to_grid_dict(args):
    outline = args.grid_outline[0] if len(args.grid_outline) == 1 else tuple(int(x) for x in args.grid_outline)
    minor_outline = (
        args.grid_minor_outline[0]
        if len(args.grid_minor_outline) == 1
        else tuple(int(x) for x in args.grid_minor_outline)
    )
    fill = args.grid_fill[0] if len(args.grid_fill) == 1 else tuple(int(x) for x in args.grid_fill)
    font_path = find_font(args.grid_font, args.grid_text_size)
    font = Font(outline, font_path, size=args.grid_text_size)
    grid_dict = {
        "lon_major": args.grid_D[0],
        "lat_major": args.grid_D[1],
        "lon_minor": args.grid_d[0],
        "lat_minor": args.grid_d[1],
        "font": font,
        "fill": fill,
        "outline": outline,
        "minor_outline": minor_outline,
        "write_text": args.grid_text,
        "width": args.grid_width,
        "lon_placement": args.grid_lon_placement,
        "lat_placement": args.grid_lat_placement,
    }
    return grid_dict


def _args_to_colorbar_kwargs(args):
    font_color = args.colorbar_text_color
    font_color = font_color[0] if len(font_color) == 1 else tuple(int(x) for x in font_color)
    font_path = find_font(args.colorbar_font, args.colorbar_text_size)
    # this actually needs an aggdraw font
    font = Font(font_color, font_path, size=args.colorbar_text_size)

    if args.colorbar_width is None or args.colorbar_height is None:
        if args.colorbar_width is not None or args.colorbar_height is not None:
            LOG.warning(
                "'--colorbar-width' and '--colorbar-height' were not both specified. " "Forcing '--colorbar-extend'."
            )
        args.colorbar_extend = True

    colorbar_kwargs = {
        "font": font,
        "tick_color": font_color,
        "cmin": args.colorbar_min,
        "cmax": args.colorbar_max,
        "align": args.colorbar_align,
        "extend": args.colorbar_extend,
        "tick_marks": args.colorbar_tick_marks,
        "minor_tick_marks": args.colorbar_minor_tick_marks,
        "title": args.colorbar_title,
        "unit": args.colorbar_units,
    }
    if args.colorbar_width:
        colorbar_kwargs["width"] = args.colorbar_width
    if args.colorbar_height:
        colorbar_kwargs["height"] = args.colorbar_height
    return colorbar_kwargs


def find_font(font_name, size):
    try:
        font = ImageFont.truetype(font_name, size)
        return font.path
    except IOError:
        font_path = get_resource_filename("polar2grid.fonts", font_name)
        if not os.path.exists(font_path):
            raise ValueError("Font path does not exist: {}".format(font_path))
        return font_path


def _process_one_image(
    input_tiff: str, output_filename: str, pycoast_options: dict, shapes_dir: str, colorbar_kwargs: dict
) -> None:
    LOG.info("Creating {} from {}".format(output_filename, input_tiff))
    img = Image.open(input_tiff)
    img_bands = img.getbands()
    num_bands = len(img_bands)
    # P = palette which we assume to be an RGBA colormap
    img = img.convert("RGBA" if num_bands in (2, 4) or "P" in img_bands else "RGB")
    if pycoast_options:
        area_id = os.path.splitext(input_tiff[0])[0]
        area_def = get_area_def_from_raster(input_tiff, area_id=area_id)
        cw = ContourWriterAGG(shapes_dir)
        cw.add_overlay_from_dict(pycoast_options, area_def, background=img)

    if colorbar_kwargs:
        colorbar_kwargs = colorbar_kwargs.copy()
        cmin = colorbar_kwargs.pop("cmin")
        cmax = colorbar_kwargs.pop("cmax")
        cmap = _get_colormap_object(input_tiff, num_bands, cmin, cmax)
        _add_colorbar_to_image(img, colormap=cmap, **colorbar_kwargs)

    img.save(output_filename)


def _get_colormap_object(input_tiff, num_bands, cmin, cmax):
    rio_ds = rasterio.open(input_tiff)
    input_dtype = np.dtype(rio_ds.meta["dtype"])
    colormap_csv = rio_ds.tags().get("colormap")
    rio_ct = _get_rio_colormap(rio_ds, 1)
    metadata = rio_ds.tags()
    cmap = _convert_table_to_cmap_or_default_bw(input_dtype, rio_ct, num_bands)
    if num_bands in (3, 4) and colormap_csv is None:
        raise ValueError("RGB and RGBA geotiffs must have a colormap " "specified with '--colorbar-colormap-file'.")
    if num_bands in (3, 4) or colormap_csv is not None:
        cmap = Colormap.from_file(colormap_csv)
    vmin, vmax = _get_colorbar_vmin_vmax(cmin, cmax, metadata, input_dtype)
    cmap = cmap.set_range(vmin, vmax, inplace=False)
    return cmap


def _convert_table_to_cmap_or_default_bw(band_dtype, band_ct, band_count):
    max_val = np.iinfo(band_dtype).max
    # if we have an alpha band then include the entire colormap
    # otherwise assume it is using 0 as a fill value
    start_idx = 1 if band_count == 1 else 0
    if band_ct is None:
        # all black colormap
        # don't assume anything about the colors or scaling of the image if the
        # user didn't provide enough information
        color_iter = ((idx / float(max_val), (0.0, 0.0, 0.0, 1.0)) for idx in range(max_val))
        color_iter = list(color_iter)
    else:
        color_iter = ((idx / float(max_val), color) for idx, color in sorted(band_ct.items())[start_idx:])
        color_iter = ((idx, tuple(x / float(max_val) for x in color)) for idx, color in color_iter)
    cmap = Colormap(*color_iter)
    return cmap


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2013 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
#
# Written by David Hoese    March 2013
# University of Wisconsin-Madison
# Space Science and Engineering Center
# 1225 West Dayton Street
# Madison, WI  53706
# david.hoese@ssec.wisc.edu
"""Script for converting lon/lat decimal values to X/Y values of the grid provided.

This script mimics the `proj` binary that comes with the PROJ.4 library, but
handles any projection string that `pyproj` can handle (like 'latlong' which the
`proj` binary does not handle).

"""

import sys

try:
    from pyproj import Proj as BaseProj
except ImportError:
    # This module is loaded by meta.py which could be all a user needs so we shouldn't fail if they don't have pyproj
    import warnings

    warnings.warn("Package 'pyproj' could not be imported. Some functionality will be missing")
    Proj = object


class Proj(BaseProj):
    def is_latlong(self):
        if hasattr(self, "crs"):
            return self.crs.is_geographic
        return super(Proj, self).is_latlong()

    def __call__(self, data1, data2, **kwargs):
        if self.is_latlong():
            return data1, data2

        return super(Proj, self).__call__(data1, data2, **kwargs)


def get_parser():
    from argparse import ArgumentParser

    parser = ArgumentParser(description="Convert latitude/longitude coordinates to X/Y values.")
    parser.add_argument(
        "-i",
        "--inverse",
        dest="inv",
        action="store_true",
        default=False,
        help="Convert X/Y values to latitude/longitude coordinates",
    )
    parser.add_argument("proj4_str", help="PROJ.4 projection string (in quotes)")
    parser.add_argument("lon_point", type=float, help="Longitude of the point to be converted (single value only)")
    parser.add_argument("lat_point", type=float, help="Latitude of the point to be converted (single value only)")
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()

    p = Proj(args.proj4_str)
    x, y = p(args.lon_point, args.lat_point, inverse=args.inv)
    print(x, y)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Test Polar2Grid-specific enhancement functions."""

import contextlib
import os
import shutil

import dask.array as da
import numpy as np
import pytest
import rasterio
import satpy
from satpy import Scene

from polar2grid.tests import TEST_ETC_DIR


@contextlib.contextmanager
def easy_cwd(new_dir=None):
    cur_dir = os.getcwd()
    try:
        if new_dir:
            os.chdir(new_dir)
        yield
    finally:
        os.chdir(cur_dir)


class TestP2GEnhancements:
    """Test p2g-specific enhancements."""

    def setup_method(self):
        """Add P2G configs to the Satpy path."""
        from polar2grid.utils.config import add_polar2grid_config_paths

        self._old_path = satpy.config.get("config_path")
        add_polar2grid_config_paths()
        # add test specific configs
        curr_path = satpy.config.get("config_path")
        satpy.config.set(config_path=[TEST_ETC_DIR] + curr_path)

    def teardown_method(self):
        """Reset Satpy config path back to the original value."""
        satpy.config.set(config_path=self._old_path)

    def test_temperature_difference(self, tmp_path, abi_l1b_c01_data_array):
        new_data_arr = abi_l1b_c01_data_array.copy()
        data = da.linspace(-10, 10, new_data_arr.size).reshape(new_data_arr.shape)
        new_data_arr.data = data
        new_data_arr.attrs["name"] = "test_temperature_difference"
        scn = Scene()
        scn["test_temperature_difference"] = new_data_arr
        scn.save_datasets(filename="{name}_{standard_name}.tif", base_dir=str(tmp_path))

        exp_fn = tmp_path / "test_temperature_difference_toa_bidirectional_reflectance.tif"
        assert os.path.isfile(exp_fn)
        with rasterio.open(exp_fn, "r") as out_ds:
            assert out_ds.count == 2
            l_data = out_ds.read(1)
            # see polar2grid/tests/etc/enhancements/generic.yaml
            flat_l_data = l_data.ravel()
            data = data.ravel().compute()
            exp_out = np.round(np.linspace(5.0, 205.0, data.size)).astype(np.uint8)
            np.testing.assert_allclose(flat_l_data, exp_out)

    @pytest.mark.parametrize(
        "ds_name",
        [
            "test_p2g_palettize",
            "test_p2g_palettize2",
            "test_p2g_palettize3",
            "test_p2g_palettize4",
            "test_p2g_colorize",
        ],
    )
    @pytest.mark.parametrize("keep_palette", [False, True])
    def test_p2g_palettize(self, keep_palette, ds_name, tmpdir, abi_l1b_c01_data_array):
        if ds_name == "test_p2g_palettize3":
            shutil.copy(os.path.join(TEST_ETC_DIR, f"{ds_name}.npy"), tmpdir)
        new_data_arr = abi_l1b_c01_data_array.copy()
        data = da.linspace(180, 280, new_data_arr.size).reshape(new_data_arr.shape)
        new_data_arr.data = data
        new_data_arr.attrs["name"] = ds_name
        scn = Scene()
        scn[ds_name] = new_data_arr
        out_fn = str(tmpdir + f"{ds_name}_{keep_palette}.tif")
        with easy_cwd(tmpdir):
            scn.save_datasets(filename=out_fn, keep_palette=keep_palette)

        with rasterio.open(out_fn, "r") as out_ds:
            is_palette = keep_palette and "palettize" in ds_name
            num_bands = 1 if is_palette else 4
            assert out_ds.count == num_bands
            if is_palette:
                assert out_ds.colormap(1) is not None



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Filter classes for making decisions based on day or night percentages."""

import logging

try:
    # Python 3.9+
    from functools import cache
except ImportError:
    from functools import lru_cache as cache

import numpy as np
from pyorbital.astronomy import sun_zenith_angle
from pyresample.spherical import SphPolygon
from xarray import DataArray

from ._base import BaseFilter
from ._utils import polygon_for_area

logger = logging.getLogger(__name__)


@cache
def _get_sunlight_coverage(area_def, start_time, sza_threshold=90.0, overpass=None):
    """Get the sunlight coverage of *area_def* at *start_time* as a value between 0 and 1."""
    if area_def.ndim == 1:
        logger.debug("Source data is 1 dimensional, will not filter based on sunlight coverage.")
        return False
    adp = polygon_for_area(area_def)
    poly = get_twilight_poly(start_time)
    if overpass is not None:
        ovp = overpass.boundary.contour_poly
        cut_area_poly = adp.intersection(ovp)
    else:
        cut_area_poly = adp

    if cut_area_poly is None:
        if not adp._is_inside(ovp):
            return 0.0
        else:
            # Should already have been taken care of in pyresample.spherical.intersection
            cut_area_poly = adp

    daylight = cut_area_poly.intersection(poly)
    if daylight is None:
        if sun_zenith_angle(start_time, *area_def.get_lonlat(0, 0)) < sza_threshold:
            return 1.0
        else:
            return 0.0
    else:
        daylight_area = daylight.area()
        total_area = adp.area()
        return daylight_area / total_area


def modpi(val, mod=np.pi):
    """Put *val* between -*mod* and *mod*."""
    return (val + mod) % (2 * mod) - mod


def get_twilight_poly(utctime):
    """Return a polygon enclosing the sunlit part of the globe at *utctime*."""
    from pyorbital import astronomy

    ra, dec = astronomy.sun_ra_dec(utctime)
    lon = modpi(ra - astronomy.gmst(utctime))
    lat = dec

    vertices = np.zeros((4, 2))

    vertices[0, :] = modpi(lon - np.pi / 2), 0
    if lat <= 0:
        vertices[1, :] = lon, np.pi / 2 + lat
        vertices[3, :] = modpi(lon + np.pi), -(np.pi / 2 + lat)
    else:
        vertices[1, :] = modpi(lon + np.pi), np.pi / 2 - lat
        vertices[3, :] = lon, -(np.pi / 2 - lat)

    vertices[2, :] = modpi(lon + np.pi / 2), 0

    return SphPolygon(vertices)


class SunlightCoverageFilter(BaseFilter):
    """Base class for filtering based on day/night coverage."""

    FILTER_MSG = "Unloading '{}' because there is not enough day/night coverage."

    def __init__(self, product_filter_criteria: dict = None, sza_threshold: float = 100.0, fraction: float = 0.1):
        """Initialize thresholds and default search criteria."""
        super().__init__(product_filter_criteria)
        self._sza_threshold = sza_threshold
        self._fraction = fraction

    def _should_be_filtered(self, sunlight_coverage):
        raise NotImplementedError("Subclass must implement the filter decision")

    def _filter_data_array(self, data_arr: DataArray, _cache: dict):
        """Check if this DataArray should be removed.

        Returns:
            True if it meets the condition and does not have enough
            day data and should therefore be removed. False otherwise
            meaning it should be kept.

        """
        if not self._matches_criteria(data_arr):
            return False

        area = data_arr.attrs["area"]
        start_time = data_arr.attrs["start_time"]
        end_time = data_arr.attrs["end_time"]
        mid_time = start_time + (end_time - start_time) / 2
        sza_threshold = self._sza_threshold
        if area not in _cache:
            slc = _get_sunlight_coverage(area, mid_time, sza_threshold)
            logger.debug("Sunlight is estimated to cover %0.2f%% of the scene.", slc * 100)
            _cache[area] = self._should_be_filtered(slc)
        return _cache[area]


class DayCoverageFilter(SunlightCoverageFilter):
    """Remove certain products when there is not enough day data."""

    FILTER_MSG = "Unloading '{}' because there is not enough day data."

    def __init__(self, product_filter_criteria: dict = None, sza_threshold: float = 100.0, day_fraction: float = 0.1):
        """Initialize thresholds and default search criteria."""
        super().__init__(product_filter_criteria, sza_threshold=sza_threshold, fraction=day_fraction)

    def _should_be_filtered(self, sunlight_coverage):
        return sunlight_coverage < self._fraction


class NightCoverageFilter(SunlightCoverageFilter):
    """Remove certain products when there is not enough night data."""

    FILTER_MSG = "Unloading '{}' because there is not enough night data."

    def __init__(self, product_filter_criteria: dict = None, sza_threshold: float = 100.0, night_fraction: float = 0.1):
        """Initialize thresholds and default search criteria."""
        super().__init__(product_filter_criteria, sza_threshold=sza_threshold, fraction=night_fraction)

    def _should_be_filtered(self, sunlight_coverage):
        return (1 - sunlight_coverage) < self._fraction



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2014 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
#     input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Functions that handle polar2grid data types."""

import logging

import numpy as np

log = logging.getLogger(__name__)

DTYPE_UINT8 = "uint1"
DTYPE_UINT16 = "uint2"
DTYPE_UINT32 = "uint4"
DTYPE_UINT64 = "uint8"
DTYPE_INT8 = "int1"
DTYPE_INT16 = "int2"
DTYPE_INT32 = "int4"
DTYPE_INT64 = "int8"
DTYPE_FLOAT32 = "real4"
DTYPE_FLOAT64 = "real8"

NUMPY_DTYPE_STRS = [
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "int8",
    "int16",
    "int32",
    "int64",
    "float32",
    "float64",
]

# Map data type to np type for conversion
str2dtype = {
    DTYPE_UINT8: np.uint8,
    DTYPE_UINT16: np.uint16,
    DTYPE_UINT32: np.uint32,
    DTYPE_UINT64: np.uint64,
    DTYPE_INT8: np.int8,
    DTYPE_INT16: np.int16,
    DTYPE_INT32: np.int32,
    DTYPE_INT64: np.int64,
    DTYPE_FLOAT32: np.float32,
    DTYPE_FLOAT64: np.float64,
}
dtype2str = dict((v, k) for k, v in str2dtype.items())

# Map data type to valid data range
# Since float32 is polar2grid's intermediate type, float32 and float64 don't
# get clipped
dtype2range = {
    DTYPE_UINT8: (0, 2**8 - 1),
    DTYPE_UINT16: (0, 2**16 - 1),
    DTYPE_UINT32: (0, 2**32 - 1),
    DTYPE_UINT64: (0, 2**64 - 1),
    DTYPE_INT8: (-1 * (2**7), 2**7 - 1),
    DTYPE_INT16: (-1 * (2**15), 2**15 - 1),
    DTYPE_INT32: (-1 * (2**31), 2**31 - 1),
    DTYPE_INT64: (-1 * (2**63), 2**63 - 1),
    DTYPE_FLOAT32: None,
    DTYPE_FLOAT64: None,
}


def normalize_dtype_string(dtype_str):
    if dtype_str in str2dtype:
        return dtype_str
    else:
        msg = "Unknown data type string '%s'" % (dtype_str,)
        log.error(msg)
        raise ValueError(msg)


def str_to_dtype(dtype_str):
    if np.issubclass_(dtype_str, np.number):
        # if they gave us a np dtype
        return dtype_str
    elif isinstance(dtype_str, str) and hasattr(np, dtype_str):
        # they gave us the np name of the dtype
        return getattr(np, dtype_str)

    try:
        return str2dtype[dtype_str]
    except KeyError:
        raise ValueError("Not a valid data type string: %s" % (dtype_str,))


def dtype_to_str(numpy_dtype):
    if isinstance(numpy_dtype, str):
        # if they gave us a string, make sure it's valid
        return normalize_dtype_string(numpy_dtype)

    try:
        return dtype2str[np.dtype(numpy_dtype).type]
    except KeyError:
        raise ValueError("Unsupported np data type: %r" % (numpy_dtype,))


def clip_to_data_type(data, data_type):
    if np.issubdtype(np.dtype(data_type), np.floating):
        return data
    info = np.iinfo(data_type)
    rmin = info.min
    rmax = info.max
    log.debug("Clipping data to a %d - %d data range" % (rmin, rmax))
    return data.clip(rmin, rmax)


def int_or_float(val):
    """Convert a string to integer or float.

    If we can't make it an integer then make it a float.

    """
    try:
        return int(val)
    except ValueError:
        return float(val)



#!/usr/bin/env python3
# encoding: utf-8
# Copyright (C) 2013-2015 Space Science and Engineering Center (SSEC),
# University of Wisconsin-Madison.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
#     input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
#
# Written by David Hoese    March 2015
# University of Wisconsin-Madison
# Space Science and Engineering Center
# 1225 West Dayton Street
# Madison, WI  53706
# david.hoese@ssec.wisc.edu
"""Utilities and accessor functions to grids and projections used in polar2grid."""
from __future__ import annotations

import logging
import os

from pyproj import CRS, Proj

from polar2grid.core.containers import GridDefinition

LOG = logging.getLogger(__name__)


def _parse_meter_degree_param(param) -> tuple[float, bool]:
    """Parse a configuration parameter that could be meters or degrees.

    Degrees are denoted with a suffix of 'deg'. Meters are denoted with a suffix of either 'm' or no suffix at all.

    :returns: (float param, True if degrees/False if meters)
    """
    is_deg = False
    if param.endswith("deg"):
        # Parameter is in degrees
        is_deg = True
        param = param[:-3]
    elif param.endswith("m"):
        # Parameter is in meters
        param = param[:-1]
    return float(param), is_deg


def get_proj4_info(proj4_str):
    parts = [x.replace("+", "") for x in proj4_str.split(" ")]
    if "no_defs" in parts:
        parts.remove("no_defs")

    proj4_dict = dict(p.split("=") if "=" in p else (p, True) for p in parts)
    _convert_numeric_parameters_to_floats(proj4_dict)
    return proj4_dict


def _convert_numeric_parameters_to_floats(proj4_dict):
    for k in ["lat_0", "lat_1", "lat_2", "lat_ts", "lat_b", "lat_t", "lon_0", "lon_1", "lon_2", "lonc", "a", "b", "es"]:
        if k in proj4_dict:
            proj4_dict[k] = float(proj4_dict[k])


def _parse_proj4_str(proj4_str, grid_name):
    # Test to make sure the proj4_str is valid in pyproj's eyes
    try:
        crs = CRS.from_proj4(proj4_str)
    except ValueError:
        LOG.error("Invalid proj4 string in '%s' : '%s'" % (grid_name, proj4_str))
        raise
    return proj4_str, crs


def parse_proj4_config_line(grid_name, parts):
    proj4_str, crs = _parse_proj4_str(parts[2], grid_name)
    parts = [None if part in ("None", "") else part for part in parts]
    is_static = any(part is None for part in parts[3:9])

    try:
        grid_width, grid_height = _parse_grid_width_and_height(grid_name, parts[3], parts[4])
        pixel_size_x, pixel_size_y = _parse_pixel_size_x_y(grid_name, parts[5], parts[6])
        grid_origin_x, grid_origin_y, grid_units = _parse_origin_x_y(grid_name, parts[7], parts[8], crs)
    except ValueError:
        LOG.error("Could not parse proj4 grid configuration: '%s'" % (grid_name,))
        raise

    if grid_width is None and pixel_size_x is None:
        LOG.error("Either grid size or pixel size must be specified for '%s'" % grid_name)
        raise ValueError("Either grid size or pixel size must be specified for '%s'" % grid_name)

    info = {}
    info["grid_kind"] = "proj4"
    info["static"] = is_static
    info["proj4_str"] = proj4_str
    info["pixel_size_x"] = pixel_size_x
    info["pixel_size_y"] = pixel_size_y
    info["grid_origin_x"] = grid_origin_x
    info["grid_origin_y"] = grid_origin_y
    info["grid_width"] = grid_width
    info["grid_height"] = grid_height
    info["grid_origin_units"] = grid_units
    return info


def _parse_grid_width_and_height(grid_name, width_part, height_part):
    grid_width = _parse_optional_config_param(width_part, int)
    grid_height = _parse_optional_config_param(height_part, int)
    if _only_one_specified(grid_width, grid_height):
        LOG.error("Both or neither grid sizes must be specified for '%s'" % grid_name)
        raise ValueError("Both or neither grid sizes must be specified for '%s'" % grid_name)
    return grid_width, grid_height


def _parse_pixel_size_x_y(grid_name, x_part, y_part):
    pixel_size_x = _parse_optional_config_param(x_part, float, (grid_name, "pixel_size"))
    pixel_size_y = _parse_optional_config_param(y_part, float, (grid_name, "pixel size"))
    if _only_one_specified(pixel_size_y, pixel_size_y):
        LOG.error("Both or neither pixel sizes must be specified for '%s'" % grid_name)
        raise ValueError("Both or neither pixel sizes must be specified for '%s'" % grid_name)
    return pixel_size_x, pixel_size_y


def _parse_origin_x_y(grid_name, x_part, y_part, crs):
    xorigin_res = _parse_optional_config_param(x_part, _parse_meter_degree_param)
    grid_origin_x, xorigin_is_deg = xorigin_res if xorigin_res is not None else (None, False)
    yorigin_res = _parse_optional_config_param(y_part, _parse_meter_degree_param)
    grid_origin_y, yorigin_is_deg = yorigin_res if yorigin_res is not None else (None, False)
    if _only_one_specified(grid_origin_x, grid_origin_x):
        LOG.error("Both or neither grid origins must be specified for '%s'" % grid_name)
        raise ValueError("Both or neither grid origins must be specified for '%s'" % grid_name)
    if xorigin_is_deg != yorigin_is_deg:
        LOG.error("Grid origin parameters must be in the same units (meters vs degrees)")
        raise ValueError("Grid origin parameters must be in the same units (meters vs degrees)")
    grid_units = "degrees" if xorigin_is_deg or crs.is_geographic else "meters"
    return grid_origin_x, grid_origin_y, grid_units


def _only_one_specified(a, b):
    only_a_specified = a is not None and b is None
    only_b_specified = a is None and b is not None
    return only_a_specified or only_b_specified


def _parse_optional_config_param(str_part, convert_func, unspecified_info=None):
    if str_part is not None:
        return convert_func(str_part)
    if unspecified_info is not None:
        grid_name, param_name = unspecified_info
        LOG.warning(f"Grid {grid_name!r} may not process properly due to unspecified {param_name}")
    return None


def parse_and_convert_proj4_config_line(grid_name, parts):
    """Return a dictionary of information for a specific PROJ.4 grid from a grid configuration line.

    ``parts`` should be every comma-separated part of the line including the ``grid_name``.
    """
    info = parse_proj4_config_line(grid_name, parts)
    _convert_origin_degrees_to_meters_if_needed(grid_name, info)
    proj4_dict = get_proj4_info(info["proj4_str"])
    info.update(**proj4_dict)
    return info


def _convert_origin_degrees_to_meters_if_needed(grid_name, info):
    p = Proj(info["proj4_str"])
    if info["grid_origin_units"] == "degrees" and not p.crs.is_geographic:
        meters_x, meters_y = p(info["grid_origin_x"], info["grid_origin_y"])
        LOG.debug(
            "Converted grid '%s' origin from (lon: %f, lat: %f) to (x: %f, y: %f)",
            grid_name,
            info["grid_origin_x"],
            info["grid_origin_y"],
            meters_x,
            meters_y,
        )
        info["grid_origin_x"] = meters_x
        info["grid_origin_y"] = meters_y
    elif info["grid_origin_units"] != "degrees" and info["grid_origin_x"] is not None and p.crs.is_geographic:
        LOG.error("Lat/Lon grid '%s' must have its origin in degrees", grid_name)
        raise ValueError("Lat/Lon grid '%s' must have its origin in degrees" % (grid_name,))


def read_grids_config_str(config_str, convert_coords=True):
    grid_information = {}
    this_configs_grids = []

    for parts in _generate_valid_parts_in_config_str(config_str):
        grid_name = parts[0]
        # Help the user out by checking if they are adding a grid more than once
        if grid_name not in this_configs_grids:
            this_configs_grids.append(grid_name)
        else:
            LOG.warning("Grid '%s' is in grid config more than once" % (grid_name,))

        grid_type = parts[1].lower()
        if grid_type != "proj4":
            LOG.error("Unknown grid type '%s' for grid '%s' in grid config" % (grid_type, grid_name))
            raise ValueError("Unknown grid type '%s' for grid '%s' in grid config" % (grid_type, grid_name))
        if convert_coords:
            grid_information[grid_name] = parse_and_convert_proj4_config_line(grid_name, parts)
        else:
            grid_information[grid_name] = parse_proj4_config_line(grid_name, parts)

    return grid_information


def _generate_valid_parts_in_config_str(config_str: str):
    for line in config_str.split("\n"):
        # Skip comments and empty lines
        if not line or line.startswith("#") or line.startswith("\n"):
            continue

        # Clean up the configuration line
        line = line.strip("\n,")
        parts = [part.strip() for part in line.split(",")]

        if len(parts) != 11 and len(parts) != 9:
            LOG.error("Grid configuration line '%s' in grid config does not have the correct format" % (line,))
            raise ValueError("Grid configuration line '%s' in grid config does not have the correct format" % (line,))
        yield parts


def read_grids_config(config_filepath, convert_coords=True):
    """Read the "grids.conf" file and create dictionaries mapping the grid name to the necessary information.

    There are two dictionaries created, one for gpd file grids and one for proj4 grids.

    Format for proj4 grids:
    grid_name,proj4,proj4_str,pixel_size_x,pixel_size_y,origin_x,origin_y,width,height

    where 'proj4' is the actual text 'proj4' to define the grid as a proj4 grid.

    """
    full_config_filepath = os.path.realpath(os.path.expanduser(config_filepath))
    with open(full_config_filepath, "r") as config_file:
        config_str = config_file.read()
        return read_grids_config_str(config_str, convert_coords=convert_coords)


class GridManager:
    """Object that holds grid information about the grids added to it."""

    grid_information: dict[str, dict] = {}

    def __init__(self, *grid_configs):
        for grid_config in grid_configs:
            LOG.debug("Loading grid configuration '%s'" % (grid_config,))
            self.add_grid_config(grid_config)

    def __contains__(self, item):
        return item in self.grid_information

    def __getitem__(self, item):
        return self.get_grid_definition(item)

    def add_grid_config(self, grid_config_filename):
        """Load a grid configuration file.

        If a ``grid_name`` was already added its information is overwritten.
        """
        grid_information = read_grids_config(grid_config_filename)
        self.grid_information.update(**grid_information)

    def add_grid_config_str(self, grid_config_line):
        grid_information = read_grids_config_str(grid_config_line)
        self.grid_information.update(**grid_information)

    def add_proj4_grid_info(self, grid_name, proj4_str, width, height, cell_width, cell_height, origin_x, origin_y):
        # Trick the parse function to think this came from a config line
        parts = (grid_name, "proj4", proj4_str, width, height, cell_width, cell_height, origin_x, origin_y)
        self.grid_information[grid_name] = parse_and_convert_proj4_config_line(grid_name, parts)

    def get_grid_definition(self, grid_name: str) -> GridDefinition:
        """Return a standard `GridDefinition` object for the specified grid.

        :returns: `GridDefinition` object, updates to this object do not affect information
                  internal to the `Cartographer`.

        """
        grid_info = self.get_grid_info(grid_name)
        return GridDefinition(
            grid_name=grid_name,
            proj4_definition=grid_info["proj4_str"],
            height=grid_info["grid_height"],
            width=grid_info["grid_width"],
            cell_width=grid_info["pixel_size_x"],
            cell_height=grid_info["pixel_size_y"],
            origin_x=grid_info["grid_origin_x"],
            origin_y=grid_info["grid_origin_y"],
        )

    def get_grid_info(self, grid_name):
        """Return a grid information dictionary about the ``grid_name`` specified.

        The information returned will always be a copy of the information
        stored internally in this object. So a change to the dictionary
        returned does NOT affect the internal information.

        :raises ValueError: if ``grid_name`` does not exist

        """
        if grid_name in self.grid_information:
            return self.grid_information[grid_name].copy()
        else:
            LOG.error("Unknown grid '%s'" % (grid_name,))
            raise ValueError("Unknown grid '%s'" % (grid_name,))



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2021 Space Science and Engineering Center (SSEC),
#  University of Wisconsin-Madison.
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This file is part of the polar2grid software package. Polar2grid takes
# satellite observation data, remaps it, and writes it to a file format for
# input into another program.
# Documentation: http://www.ssec.wisc.edu/software/polar2grid/
"""Filter classes dealing with resampling output coverage."""

import logging

try:
    # Python 3.9+
    from functools import cache
except ImportError:
    from functools import lru_cache as cache

from pyresample.spherical import SphPolygon
from xarray import DataArray

from ._base import BaseFilter
from ._utils import PRGeometry, polygon_for_area

logger = logging.getLogger(__name__)


@cache
def _get_intersection_coverage(source_polygon: SphPolygon, target_polygon: SphPolygon) -> float:
    """Get fraction of output grid that will be filled with input data."""
    intersect_polygon = source_polygon.intersection(target_polygon)
    if intersect_polygon is None:
        same_shape = source_polygon.vertices.shape == target_polygon.vertices.shape
        if same_shape and (source_polygon.vertices == target_polygon.vertices).all():
            # they are exactly the same
            return 1.0
        return 0.0
    return intersect_polygon.area() / target_polygon.area()


class ResampleCoverageFilter(BaseFilter):
    """Remove any DataArrays that would not have any results if resampled to the target area.

    This filter defaults to checking all products. Pass ``product_filter_critera=None`` to
    disable checking for all products. See :class:`BaseFilter` for other options.

    """

    def __init__(
        self,
        product_filter_criteria: dict = True,
        target_area: PRGeometry = None,
        coverage_fraction: float = None,
    ):
        """Initialize thresholds and default search criteria."""
        super().__init__(product_filter_criteria)
        if target_area is None:
            raise ValueError("'target_area' is required.")
        if coverage_fraction is None:
            raise ValueError("'coverage_fraction' is required.")
        self._target_area = target_area
        self._target_polygon = polygon_for_area(self._target_area)
        self._coverage_fraction = coverage_fraction

    def _filter_data_array(self, data_arr: DataArray, _cache: dict):
        """Check if this DataArray should be removed.

        Returns:
            True if it meets the condition and does not have enough
            day data and should therefore be removed. False otherwise
            meaning it should be kept.

        """
        if not self._matches_criteria(data_arr):
            return False
        source_area = data_arr.attrs.get("area")
        if source_area is None:
            return False
        if source_area.ndim == 1:
            logger.debug("Source data is 1 dimensional, will not filter based on resampling coverage.")
            return False

        coverage_fraction = self._get_and_cache_coverage_fraction(data_arr, source_area, _cache)
        if coverage_fraction >= self._coverage_fraction:
            logger.debug(f"Resampling found {coverage_fraction * 100:0.02f}% coverage.")
            return False
        logger.warning(
            f"Resampling found {coverage_fraction * 100:0.02f}% of the output grid "
            f"{self._target_area.area_id!r} covered. Will skip producing this product: {data_arr.attrs['name']}",
        )
        return True

    def _get_and_cache_coverage_fraction(self, source_data: DataArray, source_area: PRGeometry, cache: dict) -> float:
        platform_name = source_data.attrs.get("platform_name")
        sensor_name = source_data.attrs.get("sensor")
        if sensor_name and not isinstance(sensor_name, str):
            sensor_name = "-".join(sorted(sensor_name))
        key = (platform_name, sensor_name)
        if key in cache:
            logger.debug(f"Reusing cache coverage fraction for {key}")
            return cache[key]

        logger.debug(f"Computing coverage fraction for {key}")
        source_polygon = polygon_for_area(source_area)
        coverage_fraction = _get_intersection_coverage(source_polygon, self._target_polygon)
        if platform_name is None or sensor_name is None:
            logger.debug("Platform and sensor metadata missing from data. Won't cache coverage information.")
            return coverage_fraction
        cache[key] = coverage_fraction
        return coverage_fraction
